local M = {}
M.vehIn = {}
M.stopTicks = {}
M.mode = 'disabled'
M.options = {}
M.startPos = {}
M.oldScenario = {}
local dropPos = nil
local lastPosition = {}
local helper = require('scenario/scenariohelper')
local playerID
local gdState = 'stop'
M.vehList = {}
M.driveTrain = 0
local showEndScreen = true

local function selection(response)
	M.options = response 
	if M.options then 
		for k,v in pairs(M.options) do
			if k == "opponents" then
				v = tonumber(string.match(M.options.opponents, "%d+"))
			end
			M.options[k] = v
		end
	end
	--dump(M.options)
end
local function onVehicleDestroyed(vid)
	local veh = be:getObjectByID(vid)
	if veh then
		--print('unloaded vehicle')
		helper.queueLuaCommand(veh, 'extensions.unload("gdai")')
		helper.queueLuaCommand(veh, 'extensions.unload("gdcallback")')
		helper.queueLuaCommand(veh, 'extensions.unload("gdmain")')
	else
		log('D', 'Delete Vehicle', "Error unloading extensions")
	end
end
local function deletePrefabs()
	local objName = nil
	local obj = nil
	local arena = M.options.arena.name
	local map = M.options.fmap
	local arenaType = M.options.arena.type
	local objNames = {'derby',arena}
	for _,objName in ipairs(objNames) do
		obj = scenetree[objName]
		if obj then
			local fn = 'levels/'..map..'/prefabs/'..arenaType..'/'..objName..'.prefab'
			if FS:fileExists(fn) then
				obj:unload()
				obj:delete()
				log('D', 'scenario_scenarios.spawnPrefab', 'unloading Prefab '..objName)
			end
			be:reloadCollision()
		end
	end
end
local function reset(state)
	print('reset'..state)
	if gdState == 'stop' then return end  -- need this so it does not break other scenarios
	for k,v in ipairs(M.vehIn) do
		if v.vid ~= playerID then
			local veh = be:getObjectByID(v.vid)
			if veh then
				local mode = 'stop'
				helper.queueLuaCommand(veh, 'gdai.setState({mode = "'..mode..'"})')
			end
		end
	end
	deletePrefabs()
	M.vehIn = {}
	M.driveTrain = 0
	M.stopTicks = {}
	dropPos = nil
	lastPosition = {}
	if state == 1 then
		M.vehList = {}
		gdState = 'stop'
	elseif state == 2 then
		gdState = 'heat'
	else
		gdState = 'pre-running'
	end
end
local function onScenarioRestarted()
	reset(1)
end
function getTableSize(t)
    local count = 0
    for _, __ in pairs(t) do
        count = count + 1
    end
    return count
end
local function endScreen() 
	M.mode = 'disabled'
	showEndScreen = false
	if bullettime.get() > 1/8 then bullettime.set(1/8) end
	local endData = {}
	 local buttonsTable = {}
	buttonsTable = {{label='RETRY', cmd='scenario_scenarios.uiEventRetry()'}, {label='FREEROAM', cmd='scenario_scenarios.uiEventFreeRoam()'}, {label='MENU', cmd='openMenu'}, {label='SCENARIOS', cmd='openScenarios'}}
	endData.buttons = buttonsTable

	local post = scenario_gdstats.post
	local drivers = {}

	local map = M.options.fmap
	local track = M.options.arena.name
	local stats = post.stats
	local count = getTableSize(stats)
	local fnJson = 'levels/'..map..'/prefabs/purse.json'
	local purse = math.random(1,10)
	local breakdown = {}
	if FS:fileExists(fnJson) then
		local file = readJsonFile(fnJson)
		purse = file.value[track] or purse
		breakdown = file.breakdown
	end
	purse = purse * 1000
	local strCount = tostring(count)
	local money =  0
	for _,v in pairs(stats) do
		for k,i in pairs(breakdown[strCount]) do
			if k == v.place then
				money =  purse * ((tonumber(i) * 10) / 100)
			end
		end
		if M.options.arena.type == 'Racing' then
			table.insert(drivers, {place = v.place, img = "modules/gdcommon/drivers/driver("..v.driverNumber..").png", driverName = v.driverName, driverNumber = v.driverNumber, points =  v.points, model = v.model, laps = v.lap, money = money, current = v.current, out = v.out })		
		else
			table.insert(drivers, {place = v.place, img = "modules/gdcommon/drivers/driver("..v.driverNumber..").png", driverName = v.driverName, driverNumber = v.driverNumber, points =  v.points, model = v.model, knockouts = #v.knockouts, money = money, current = v.current, out = v.out })		
		end
	end
	
	endData.options = M.options
	
	guihooks.trigger('ChangeState', {state = 'gdEndScreen', params = { endData = endData, headers = post.headers, drivers = drivers}});

end
local function onRaceTick()
	if gdState == 'finished' then
		endScreen()
		gdState = 'done'
	end
end
local function onRaceResult(finished)
	if gdState == 'running' or finished then
		local heat = M.options.heat
		if heat == true and scenario_gdheats.vehHeat ~= 0 or heat == false then
			local vehIn = #M.vehIn
			local result	
			if vehIn == 1 then
				scenario_scenarios.endRace()
				gdState = 'finished'
			elseif finished then
				gdState = 'done'
				--print('show end screen')
				endScreen()	
			end
		end
	end
end
local function breakStick(vid)
	local veh = be:getObjectByID(vid)
	if veh then
		local vehData = extensions.core_vehicle_manager.getVehicleData(vid)
		if not vehData then return end
		for _, flexbody in pairs(vehData.vdata.flexbodies) do
			if type(flexbody.mesh) == "string" and (string.find(flexbody.mesh, "stickbreak") ~= nil) then
				--if flexbody.fid then
					--local f = veh:getFlexmesh(flexbody.fid)
					--if f then
						veh:setMeshAlpha(0, flexbody.mesh, false)
						--f:setAlpha(0)
					--end
				--end
			end
		end
	end
end

local function addStopTicks(dt)
	local maxStopTicks =  tonumber(string.match(M.options.cdtimer, "%d+")) or 60
	for k,v in ipairs(M.vehIn) do
		local veh = be:getObjectByID(v.vid)
		if veh then
			local pos = veh:getPosition()
			--if not M.stopTicks[v.vid] then M.stopTicks[v.vid] = 0 end 
			local stopTicks = 0--M.stopTicks[v.vid] 
			--if stopTicks ~= nil then
			local driverInfo = scenario_gdstats.getDriverInfo(v.vid)
				if driverInfo then
					if M.mode == 'derby' then
						stopTicks = driverInfo.tolHit
					else
						stopTicks = driverInfo.stopTicks
						local momentum = 0 	
						momentum = driverInfo.momentum
						if momentum > 0.4 then
							stopTicks = 0
						else
							stopTicks = stopTicks + dt
						end
					end
				end
				if not dropPos then
					dropPos = pos.z
				end
			
				if pos.z < dropPos - 2 and M.options.arena.name == 'Free Fall' then
					stopTicks = maxStopTicks + 1
				end
				--If they reach the max stop ticks then they are out
				if stopTicks >= maxStopTicks  then	
					breakStick(v.vid)
					scenario_gdstats.updateVehStats(v.vid)
					if v.vid == playerID then 
						local msg = 'You have been eliminated!'
						guihooks.trigger('Message', {ttl = 4, msg = msg, category = "content", icon = "report"})
					else
						local msg = '#'..v.driverNumber..' - '..v.driverName..' was eliminated!'
						guihooks.trigger('Message', {ttl = 4, msg = msg, category = "action", icon = "announcement"})
					end
					table.remove(M.vehIn, k)
					helper.queueLuaCommand(veh, 'controller.setFreeze('..tostring(1) ..')')
				else
					if M.mode == 'derby' then
						if v.vid == playerID and stopTicks > maxStopTicks/2 then
							if stopTicks % 1 == 0 then
								local TimeLeft = maxStopTicks - math.modf(stopTicks)
									if lastPosition[v.vid] == nil then lastPosition[v.vid] = pos end
									local distance =  math.abs((pos - lastPosition[v.vid]):len())
									local msg = tostring('No Sandbagging! You must hit someone in '..TimeLeft..' seconds.')
									if distance < 0.4 then
										msg = tostring('You need make a hit or you will be eliminated!  You have ' ..TimeLeft.. ' seconds left.')
									end									
								guihooks.trigger('Message', {ttl = 1, msg = msg, category = "fill", icon = "flag"})
							end
						end
					else
						if v.vid == playerID and stopTicks > maxStopTicks/2 then
							if stopTicks % 1 == 0 then
								local TimeLeft = maxStopTicks - math.modf(stopTicks)
								msg = tostring('You will be eliminated in ' ..TimeLeft.. ' seconds.')									
								guihooks.trigger('Message', {ttl = 1, msg = msg, category = "fill", icon = "flag"})
							end
						end
						scenario_gdstats.updateStopTicks(stopTicks,v.vid)
					end
				end
				--M.stopTicks[v.vid] = stopTicks			
			--end
		end
	end
end
local function startScenario()
-------------------------------------------------------added for beams
	--scenario_gdbeams.getModelFiles()
------------------------------------------------	
	gdState =  'loading'
	local mode = M.options.arena.type
	if mode == 'Derby' then 
		M.mode = 'derby' 
	else
		M.mode = 'race'
	end	
	scenario_gdsetup.setScenario()
	gdState = 'start'
	--	onSpineAnimationFinished_callback = true
	
	scenario_gdsetup.resetTimer()
	scenario_scenarios.onScenarioUIReady('play')  --One last call to let everything know we can start
		--extensions.hook('onScenarioChange', scenario)
end
local function onScenarioChange()
	local scenario = scenario_scenarios.getScenario()
	if not scenario then return end
	print(scenario.state)
	print(gdState)
	--if scenario.state == 'running' and gdState == 'paused' then
		--gdState = 'stop'
		--return
	--end
	if scenario.scenarioName == 'gdcommon' then
		if scenario.state == 'pre-start' or scenario.state =='pre-running' then
			if gdState == 'stop' then
				gdState = 'pre-running'
			end
			if gdState == 'pre-running' and scenario.state == 'pre-running' then
				core_vehicles.removeAll()
			end
		end
		if scenario.state == 'running' and gdState == 'pre-running' then
			if M.options.fmap then 
				if M.options.fmap == getMissionFilename():match('levels/([%w|_|%-|%s]+)/') then
					gdState = 'pre-start'
				end
			end
	--testing a loading screen
		--guihooks.trigger('ShowApps', false)
		--guihooks.trigger('ChangeState', {state = 'gdloading'})
		
		end
		if scenario.state == 'running'and gdState == 'pre-start' then
			startScenario()
			return
		end
		if scenario.state == 'post' and gdState ~= 'stop' then
			--if showEndScreen == true then 
				endScreen()
			--else 
				reset(1) 
			--end
		end
		if scenario.state == 'restart' and gdState ~= 'stop' then
			reset(1)
		end
	else 
		gdState = 'stop'
	end
end
local function freezeAll(state)
	for _, v in pairs(M.vehIn) do
		local bo = be:getObjectByID(v.vid)
		if bo then
			helper.queueLuaCommand(bo, 'controller.setFreeze('..tostring(state) ..')')
		end
    end
end
local function onCountdownStarted()
	if gdState == 'start' then
		local currentUserVehicle = be:getPlayerVehicle(0)
		local currentVehID = currentUserVehicle:getID()
		if currentVehID ~= playerID then
			local veh = be:getObjectByID(playerID)
			if veh then
				be:enterVehicle(0, veh.obj)
				commands.setCameraPlayer()
			end
		end
		gdState = 'countdown'
	end
end
local function setAi()
	if scenario_scenarios then
		gdState = 'running'
		for _,v in ipairs(M.vehIn) do
			local veh = be:getObjectByID(v.vid)
			--helper.queueLuaCommandByName(vName, 'gdai.setState({mode = "'..mode..'", debugMode = "speeds"})')
			if veh then
					--veh:queueLuaCommand([[
					--print("The rails")
					--dump(v.data.rails)
					--print("The slideNodes")
					--dump(v.data.slidenodes)
					--gdbeams.nodeInfo()
					--]])
				if M.mode == 'derby' then
					if v.vid ~= playerID then
						
						helper.queueLuaCommand(veh, 'gdai.setState({mode = "'..M.mode..'"})')
					end
				else
					scenario_gdsetup.setAiRacing(v.vid, playerID)
				end
			else
				log('D', 'logTag', 'veh not found on Countdown End', vid)
			end
			freezeAll(0)	
		end
	end
end
local function onCountdownEnded()
	if gdState == 'countdown' then
		if scenario_gdsetup then
			scenario_gdsetup.initScenario(playerID)
			setAi()
		else
			gdState = 'stop'
		end
	end
end
local function onPhysicsUnpaused()
	if gdState == 'paused' then
		if scenario_gdvehicles then
			scenario_gdvehicles.loadExtensions(2)
			setAi()
		else
			gdState = 'stop'
		end
	end
end
local function onPhysicsPaused()
	if gdState == 'running'  then
		gdState = 'paused'
		for k,v in ipairs(M.vehIn) do
			if v.vid ~= playerID then
				local veh = be:getObjectByID(v.vid)
				if veh then
					local mode = 'stop'
					helper.queueLuaCommand(veh, 'gdai.setState({mode = "'..mode..'"})')
				end
			end
		end
	end
end
local function setPlayer(renew)
	local vehPick =	M.options.vehicles.name
	--print(vehPick)
	if vehPick == 'Auto Select' and renew == 1 then
		local vehIn = #M.vehIn
		local player = math.random(1,vehIn)
		playerID = M.vehIn[player].vid
	elseif renew == 2 then
		local veh = be:getPlayerVehicle(0)
		if veh then
			playerID = veh:getID()
		else
			playerID = M.vehIn[1].vid
		end
	elseif renew == 3 then
		local playerName
		--Get Steam player name 
		if Steam and Steam.isWorking and Steam.accountLoggedIn then
			playerName = Steam.playerName
			--print("steam username: " .. Steam.playerName)
		end
		local found = false
		for k,v in ipairs(M.vehIn) do
			if playerName and v.driverName == playerName then
				playerID = M.vehIn[k].vid
				found = true
				break
			end
			if v.vid == playerID then
				found = true
			end
		end
			if found == false then
				local vehIn = #M.vehIn
				local player = math.random(1,vehIn)
				playerID = M.vehIn[player].vid
			end
	else
		playerID = M.vehIn[1].vid
			--print(playerID)
	end	
	local veh = be:getObjectByID(playerID)
	if veh then
		be:enterVehicle(0, veh.obj)
		commands.setCameraPlayer()
	end
	M.options.vehicles.name = 'Auto Select'
	return playerID
end
--This function is called on the map/level being loaded. 
------------------------------------------------
local function onClientStartMission(mission)
	if gdState == 'mapupdate' then
		gdState = 'loadmap'
	end
end
local function loadMap()
	if gdState == 'map' then
		scenario_scenarios.changeState('pre-start')
		gdmain.loadgdscenario()
		gdState = 'mapupdate'
	end
	if scenario_scenarios then
		local scenario = scenario_scenarios.getScenario()
		if not scenario then return end
		-------------------------------------------------------
		--Checks if the map is the same as the selected map.  If not reload the scenario.-------------------
		-----------------------------------------------------
		if scenario.state == 'running' then
			if M.options.fmap then 
				local map = getMissionFilename():match('levels/([%w|_|%-|%s]+)/')
				if not map then
					map = core_levels.getLevelName(getMissionFilename())
				end
				
				if M.options.fmap ~= map then
					gdState = 'map'
				end
			end
		end
		--------------------------------------------
		--This is for second time it loads the scenario to remove the start menu--
		----------------------------------------------
		if gdState == 'loadmap' then
			scenario_scenarios.changeState('running')
			guihooks.trigger('ChangeState', 'menu')
			gdState = 'pre-start'
			extensions.hook('onScenarioChange', scenario)
		end
	end
end
local function runScenario(dt)
	if scenario_gdsettings then
		scenario_gdsettings.update()
	end
	addStopTicks(dt)
	if M.mode ~= 'race' then
		if scenario_gdheats then
			scenario_gdheats.runHeats(dt)
		end
	end
	local finished
	if scenario_gdstats then
		finished = scenario_gdstats.update(playerID)	
	else
		finished = true
	end
	onRaceResult(finished)
end
local function updateGFX(dt,dtSim)
	if not scenario_scenarios then
		if gdState ~= 'stop' then
			gdState = 'stop'
		end
		return
	end
	if gdState == 'running' then
		runScenario(dt)
	elseif gdState == 'heat' then
		if scenario_gdheats then
			scenario_gdheats.runHeats(dt)
		end
	else
		if gdState ~= 'stop' then
			loadMap()
		end
	end
end

M.endScreen = endScreen

--M.onSpineAnimationFinished = onSpineAnimationFinished
M.pauseScenario = pauseScenario
M.getDriver = getDriver
M.setPlayer = setPlayer
M.selection = selection
M.onVehicleDestroyed = onVehicleDestroyed
M.onCountdownEnded = onCountdownEnded
M.onCountdownStarted = onCountdownStarted
M.onScenarioChange = onScenarioChange
M.onPhysicsPaused = onPhysicsPaused
M.updateGFX = updateGFX
M.reset = reset
M.onRaceTick = onRaceTick
M.startScenario = startScenario
M.onPhysicsUnpaused = onPhysicsUnpaused
M.onPhysicsPaused = onPhysicsPaused
M.onScenarioRestarted = onScenarioRestarted
M.onClientStartMission = onClientStartMission 
return M