-------------------------------------------------------------------------------------
-- Derby Lua Script
-- By Bidwars
-------------------------------------------------------------------------------------

local M = {}

local helper = require('scenario/scenariohelper')
--local raceMarker = require("scenario/race_marker")
local waypoint = {}
--local inputActionFilter = extensions.core_input_actionFilter

local function updateScenarioData(scenario)
	local newScenario = scenario
	local copyScenario = scenario_scenarios
	copyScenario['setScenario'] =  function(sc) 
		scenario = sc
		end  
	scenario_scenarios = copyScenario
	scenario_scenarios.setScenario(newScenario)
end
local function updateScenarioVehicles()
	local vehicles = scenetree.findClassObjects('BeamNGVehicle')
	if vehicles then
		for k, vecName in ipairs(vehicles) do
			local to = scenetree.findObject(vecName)
			if to then
				local ScenarioObjectsGroup = scenetree.findObject('ScenarioObjectsGroup')
				if ScenarioObjectsGroup then
					scenetree.ScenarioObjectsGroup:addObject(to.obj)
				end
			end
		end
	end
end
local function setScenario()
	local arenaGroup = gdscenarios.options.arena.type
	local arena = gdscenarios.options.arena.name
	local mapName = gdscenarios.options.fmap
	local scenario = scenario_scenarios.getScenario()
	-- try to load some defaults
	local directory = 'prefabs/'
	scenario.prefabs = {}
	local trackName = 'wp_'
	--local objName
	local fn = "levels/"..mapName.."/"..directory.."/"..arenaGroup.."/"..arena..".prefab"
	if FS:fileExists(fn) then
		--print(fn)
		table.insert(scenario.prefabs, fn)
	else
		fn = "levels/"..mapName.."/"..directory.."" .. mapName..".prefab"
		if FS:fileExists(fn) then
		--print(fn)
			table.insert(scenario.prefabs, fn)
		end
	end
	
	guihooks.trigger('DerbyStatsUpdate', nil)
	if arenaGroup == 'Derby' then
		scenario.lapCount = 1
	else
		trackName = 'road'
		local fnJson = "levels/"..mapName.."/"..directory.."/"..arenaGroup.."/tracks.json"
		if FS:fileExists(fnJson) then
			local trackFile = jsonReadFile(fnJson)
			if trackFile then
				trackName = trackFile[arena]
			end
		end
		directory = 'prefabs/waypoints/'
		fn = scenario.mission..""..directory.."" .. arena ..".prefab"
		if FS:fileExists(fn) then
			--print(fn)
			table.insert(scenario.prefabs, fn)
		end
	end
	--dump(scenario.prefabs)
	updateScenarioData(scenario) --saves the information to the scenario
	scenario = scenario_scenarios.getScenario()
	scenario_scenarios.onClientStartMission(scenario.mission)  --create the prefabs spawn prefabs
	scenario_gdvehicles.createVehicles()
	local waypointsTable = scenetree.findClassObjects('BeamNGWaypoint') --find the new waypoints we just spawned from prefabs
	--dump(waypointsTable)
	scenario.lapConfig = {}  --set the scenario lap config
	local finalwp = nil
	for k, nodeName in ipairs(waypointsTable) do
		if trackName and string.startswith( nodeName, trackName ) then
			local indexk = tonumber(string.match(nodeName, "%d+"))
			if not indexk then
				finalwp = nodeName
				log('D', 'scenario', tostring(k) .. ' = ' .. tostring(nodeName))
			else
				scenario.lapConfig[indexk] = nodeName
				log('D', 'scenario', tostring(k) .. ' = ' .. tostring(nodeName))
			end
		end
	end
	if finalwp then
		table.insert(scenario.lapConfig, finalwp)
	end
	core_gamestate.setGameState('scenario', scenario.uilayout or 'scenario', 'freeroam')
	map.assureLoad()  
	map.load()
end
local function setAiRacing(vid, playerID)
	local scenario = scenario_scenarios.getScenario()
	local arg = {}
	local veh = be:getObjectByID(vid)
	local vName = veh:getName()
	if vName then 
		local aggression = 1 + ((math.random(1, 10)) * .1)
		arg = {
			wpTargetList = waypoint,
			noOfLaps = scenario.lapCount or gdscenarios.options.laps , 
			wpSpeeds = {},
			aggression = aggression or 1.4,
			aggressionMode = '', 
			resetLearning = 'false',
			routeSpeed = 0,
			routeSpeedMode = 'off',
			driveInLane = 'off',
			avoidCars = 'on',
			}	
			--helper.queueLuaCommandByName(vName, 'gdai.setState({mode = "'..mode..'", debugMode = "speeds"})')
			if vid ~= playerID then
				helper.queueLuaCommandByName(vName, 'gdai.driveUsingPath({wpTargetList = '..serialize(arg.wpTargetList)..', routeSpeed = '..arg.routeSpeed..', routeSpeedMode = "'..arg.routeSpeedMode..'", driveInLane = "'..arg.driveInLane..'", wpSpeeds = '..serialize(arg.wpSpeeds)..', noOfLaps = '..arg.noOfLaps..', aggression = '..arg.aggression..', aggressionMode = "'..arg.aggressionMode..'", resetLearning = '..arg.resetLearning..', avoidCars = '..arg.avoidCars..'})')		
				if scenario_scenarios then		
					scenario_scenarios.updateVehicleAiState(vName, {mode = 'ai'})
				end
			else
				if scenario_scenarios then		
					scenario_scenarios.updateVehicleAiState(vName, {mode = 'ai'})
				end
			end
	end
end
local function initScenario(playerID)
	local scenario = scenario_scenarios.getScenario()
	if scenario then
		--local arenaGroup = gdscenarios.options.arena.type
		scenario.endUIDelayTime = math.huge
		updateScenarioData(scenario)
		--if arenaGroup ~= 'Derby' then
			scenario.state = 'pre-start'
			local veh = be:getObjectByID(playerID)	
			if veh then 
				local vName = veh:getName()
				veh.obj.playerUsable = true
				scenario.playerUsableVehicles = {vName}
				scenario.vehicles = {[''..vName..''] = {playerUsable = true, startFocus = true, driver = {}}, ['*'] = { playerUsable = false,driver = {}}}
			end
			scenario.track = {}
			scenario.track = {customMarker = 'gdcommonMarker'}
			scenario.lapCount = gdscenarios.options.laps or 10
			waypoint = scenario.lapConfig
			table.insert(waypoint, scenario.lapConfig[1])
			--scenario.endUIDelayTime = math.huge
			scenario.BranchLapConfig = deepcopy(scenario.lapConfig)
			scenario.initialLapConfig = deepcopy(scenario.lapConfig)
			--scenario.statistics = {disableEndUI = true}
			scenario.whiteListActions = {"switch_next_vehicle", "switch_previous_vehicle", "loadHome", 
							"saveHome", "recover_vehicle", "reload_vehicle", "vehicle_selector", "parts_selector", "dropPlayerAtCamera", 
							"nodegrabberRender", "slower_motion", "faster_motion", "realtime_motion", "slow_motion"}
			scenario.blackListActions =  {}
			scenario.preStartStage = 1
			updateScenarioData(scenario)
			scenario_scenarios.onPreRender(0.00,0.00)  -- set the scenario back to pre-start state.  This sets up the scenario with all the correct information
			scenario_scenarios.changeState('pre-running') --Set the waypoints 
			core_checkpoints.onScenarioRestarted(scenario)
			scenario.state = 'running' --set the scenario.state back to running
			scenario.raceState = 'racing'  --sets the timer back to racing again.
			updateScenarioData(scenario)
			--dump(scenario)
			guihooks.trigger('ChangeState', {state ='menu'}) --close the scenario menu
			extensions.hook('onScenarioChange', scenario) --One last call to let everything know we can start
		--end
		core_gamestate.setGameState('scenario', scenario.uilayout or 'scenario', 'freeroam')
	end	
end
local function resetTimer()
	local scenario = scenario_scenarios.getScenario()
	if scenario then
		scenario.showCountdown = true
		scenario.raceState = 'countdown'
		scenario.timer = 0
		scenario.countDownTime = 3.5
		scenario.countDownShowed = false
		scenario.timerActive = true
		updateScenarioData()
	end
end
local function onScenarioRestarted()
	local scenario = scenario_scenarios.getScenario()
	if scenario then
		scenario.playerUsableVehicles = {}
		scenario.vehicles = {}
		scenario.vehicleIdToName = {}
		scenario.vehicleNameToId = {}
		scenario.startingTransforms = {}
		scenario.focusSlot = nil
		scenario.nodes = {}
		scenario.lapConfig = {}
		scenario.branchLapConfig = {}
		scenario.initialLapConfig = {}
		updateScenarioData()
	end
	scenario_waypoints.onScenarioRestarted(scenario)
	waypoint = {}
end

M.resetTimer = resetTimer
M.updateScenarioData = updateScenarioData
M.setAiRacing = setAiRacing
M.onScenarioRestarted = onScenarioRestarted
M.setScenario = setScenario
M.initScenario = initScenario

return M