local C = {}
local basePrefix = "base_marker_"
local sidesPrefix = "cylinder_marker_"
local distantPrefix = "distant_marker_"
local baseShape = "art/shapes/interface/sideMarker/checkpoint_curve_base.dae"
local sideShape =  "art/shapes/interface/single_faded_column.dae"

local modeInfos = {
  default = {
    color = {0, 0, 0},
    baseColor = {0, 0, 0},
  },
  next = {
    color = {0.0, 0.0 , 0.0},
    baseColor = {0, 0, 0},
  },
  start = {
    color = {0.4, 1, 0.2},
    baseColor = {1, 1, 1},
  },
  lap = {
    color = {0.4, 1, 0.2},
    baseColor = {1, 1, 1},
  },
  final = {
    color = {0.1, 0.3, 1},
    baseColor = {1, 1, 1},
  },
  branch = {
    color = {0, 0, 0},
    baseColor = {0, 0, 0},
  },
  hidden = {
    color = {0, 0, 0},
    baseColor = {0, 0, 0},
  }
}


local closeFadeNear = 5
local closeFadeFar = 30
local distantFadeNear = 50
local distantFadeFar = 75

local distanceScale = 4
local sideToDistantRatio = distanceScale*0.1/1.215

local function inverseLerp(min, max, value)
 if math.abs(max - min) < 1e-30 then return min end
 return (value - min) / (max - min)
end

-- todo: replace this by a HSV-lerp if blending with non-gray colors
local function lerpColor(a,b,t)
  return {lerp(a[1],b[1],t),lerp(a[2],b[2],t),lerp(a[3],b[3],t)}
end

-- called when this object is created. initialize variables here (but dont spawn objects)
function C:init(id)
  self.id = id
  self.visible = false

  self.pos = nil
  self.radius = nil
  self.color = nil

  self.colorTimer = 0
  self.colorLerpDuration = 0.3

  self.modeInfos = deepcopy(modeInfos)

  self.base = nil
  self.sides = nil
  self.distant = nil
  self.normal = nil

  self.mode = 'hidden'
  self.oldMode = 'hidden'
end

-- called every frame to update the visuals.
function C:update(dt, dtSim)
	--[[
  self.colorTimer = self.colorTimer + dt
  if self.colorTimer >= self.colorLerpDuration then
    if self.mode == 'hidden' then
      self:hide()
    end
  end
  if not self.visible then return end

  local playerPosition = vec3(0,0,0)
  playerPosition:set(getCameraPosition())

  local distanceFromMarker = self.pos:distance(playerPosition)

  local t = clamp(self.colorTimer / self.colorLerpDuration,0,1)
  local color = lerpColor(self.modeInfos[self.oldMode or 'default'].color, self.modeInfos[self.mode or 'default'].color, t)
  local colorBase = lerpColor(self.modeInfos[self.oldMode or 'default'].baseColor, self.modeInfos[self.mode or 'default'].baseColor, t)
  self.currentColor = ColorF(color[1],color[2],color[3],1)
  self.currentColor.a = clamp(inverseLerp(1,50,distanceFromMarker),0,0.75) + 0.25

  local normal = self.normal and self.normal or (playerPosition-self.pos):normalized()
  local rot = quatFromDir(normal:z0()):toTorqueQuat()
  if distanceFromMarker > self.radius*1.5 then
    self.side = normal:cross(vec3(0,0,1))
  end
  if self.base then
    self.base.instanceColor = ColorF(colorBase[1],colorBase[2],colorBase[3],self.currentColor.a * 0.5):asLinear4F()
    if distanceFromMarker > self.radius*1.5 then
      self.base:setField('rotation', 0, rot.x .. ' ' .. rot.y .. ' ' .. rot.z .. ' ' .. rot.w)
    end
    self.base:updateInstanceRenderData()
  end
  local sideRadius = clamp(inverseLerp(3,60,distanceFromMarker),0.125,5)
  local sideHight = clamp(inverseLerp(60,180,distanceFromMarker),0,10)+1
  if self.left then
    self.left.instanceColor = self.currentColor:asLinear4F()
    self.left:setPosition(vec3(self.pos - self.side * self.radius + vec3(0,0,-10)):toPoint3F())
    self.left:updateInstanceRenderData()
    self.left:setScale(Point3F(sideRadius,sideRadius,sideHight))
  end
  if self.right then
    self.right.instanceColor = self.currentColor:asLinear4F()
    self.right:setPosition(vec3(self.pos + self.side * self.radius + vec3(0,0,-10)):toPoint3F())
    self.right:updateInstanceRenderData()
    self.right:setScale(Point3F(sideRadius,sideRadius,sideHight))
  end
  --]]
end

-- setting it to represent checkpoints. mode can be:
-- default (red, "normal" checkpoint)
-- branch (yellow, for branching paths)
-- next (black, the one after the current checkpoint)
-- lap (green, last cp in non-last lap)
-- finish (blue, last cp in last lap)
-- start (green, first cp when using rolling start)
function C:setToCheckpoint(wp)

  self.pos = vec3(wp.pos)
  self.radius = wp.radius
  self.normal = wp.normal and vec3(wp.normal) or nil
  self.side = (self.normal or vec3(1,0,0)):cross(vec3(0,0,1))

  if self.base then
    self.base:setPosition(vec3(self.pos):toPoint3F())
    self.base:setScale(Point3F(self.radius, self.radius, self.radius))
  end
  if self.left then
    self.left:setPosition(vec3(self.pos - self.side * self.radius):toPoint3F())
    self.left:setScale(Point3F(1,1,1))
  end
  if self.right then
    self.right:setPosition(vec3(self.pos + self.side * self.radius):toPoint3F())
    self.right:setScale(Point3F(1,1,1))
  end

end

function C:setMode(mode)
	mode = 'hidden'
  if mode ~= 'hidden' then
    self:show()
  end
  self.oldMode = self.mode
  self.mode = mode
  self.colorTimer = 0

  self:update(0,0)
end

-- visibility management
function C:setVisibility(v)
  self.visible = v
  if self.base then
    self.base.hidden = not v
  end
  if self.left then
    self.left.hidden = not v
  end
  if self.right then
    self.right.hidden = not v
  end
end

function C:hide()
  self.newColor = modeInfos['hidden'].color
  self.oldColor = modeInfos['hidden'].color
  self:setVisibility(false)
end
function C:show() self:setVisibility(true)  end

-- marker management
function C:createObject(shapeName, objectName)
  local marker =  createObject('TSStatic')
  marker:setField('shapeName', 0, shapeName)
  marker:setPosition(Point3F(0, 0, 0))
  marker.scale = Point3F(1, 1, 1)
  marker:setField('rotation', 0, '1 0 0 0')
  marker.useInstanceRenderData = true
  marker:setField('instanceColor', 0, '1 1 1 1')
  marker.canSave = false
  marker.hidden = true
  marker:registerObject(objectName)

  local scenarioObjectsGroup = scenetree.ScenarioObjectsGroup
  if scenarioObjectsGroup then
    scenarioObjectsGroup:addObject(marker)
  end

  return marker
end

-- creates neccesary objects
function C:createMarkers()
  if not self.base then
    self.base = self:createObject(baseShape,basePrefix..self.id)
  end
  if not self.left then
    self.left = self:createObject(sideShape,sidesPrefix.."left"..self.id)
  end
  if not self.right then
    self.right = self:createObject(sideShape,sidesPrefix.."right"..self.id)
  end
end

-- destorys/cleans up all objects created by this
function C:clearMarkers()
		--[[
 if self.base then
    if self.base.obj then
      self.base:delete()
    end
    self.base = nil
  end
  if self.left then
    if self.left.obj then
      self.left:delete()
    end
    self.left = nil
  end
  if self.right then
    if self.right.obj then
      self.right:delete()
    end
    self.right = nil
  end
--]]
end

return function(...)
  local o = {}
  setmetatable(o, C)
  C.__index = C
  o:init(...)
  return o
end