-------------------------------------------------------------------------------------
-- Derby Lua Script
-- By Bidwars
-------------------------------------------------------------------------------------

-- [[ STORE FREQUENTLY USED FUNCTIONS IN UPVALUES ]] --
local max = math.max
local min = math.min
local sin = math.sin
local asin = math.asin
local pi = math.pi
local abs = math.abs
local sqrt = math.sqrt
local floor = math.floor
local tableInsert = table.insert
local tableRemove = table.remove
local strFormat = string.format
---------------------------------
local scriptai = nil

local M = {}

M.mode = 'disabled' -- this is the main mode
M.manualTargetName = nil
M.debugMode = 'off'
M.targetObjectID = -1
M.speedMode = nil
M.routeSpeed = nil
M.extAggression = 1
M.cutOffDrivability = 0

-- [[ Simulation time step]] --
local dt

-- [[ ENVIRONMENT VARIABLES ]] --
local g = abs(obj:getGravity())
local gravityDir = vec3(0, 0, -1)
local gravityVec = gravityDir * g
----------------------------------

-- [[ PERFORMANCE RELATED ]] --
local aggression = 1
local aggressionMode
--------------------------------------------------------------------------

-- [[ AI DATA: POSITION, CONTROL, STATE ]] --
local aiPos = vec3(obj:getFrontPosition())
--aiPos.z = BeamEngine:getSurfaceHeightBelow(aiPos)
local aiDirVec = vec3(obj:getDirectionVector())
local prevDirVec = vec3(aiDirVec)
local aiUpVec = vec3(obj:getDirectionVectorUp())
local aiRightVec = vec3(0, 0, 0)
local aiVel = vec3(obj:getVelocity())
local aiSpeed = aiVel:length()
local targetSpeedSmoother = nil -- = newTemporalSmoothingNonLinear(math.huge, 0.2)
local aiWidth -- this comes back zero at this point
local aiLength -- this comes back zero at this point
local aiDeviation = 0
local aiDeviationSmoother = newTemporalSmoothing(1)
local aiCannotMoveTime = 0
local aiForceGoFrontTime = 0
local staticFrictionCoef = 1
local threewayturn = {state = 0, speedDifInt = 0}

local forces = {}

local lastCommand = {steering = 0, throttle = 0, brake = 0, parkingbrake = 0}

local driveInLaneFlag = false
local internalState = 'onroad'

local validateInput = nop
------------------------------

-- [[ CRASH DETECTION ]] --
--added crash.pos for derby..-----------------------------------------
local crash = {time = 0, manoeuvre = 0, dir = nil, pos = nil}

-- [[ OPPONENT DATA ]] --
local player
local plPrevVel
local chaseData = {suspectState = nil, suspectStoppedTimer = 0, distanceToSuspect = nil, playerSeg = nil, playerXnormOnSeg = nil, playerRoad = nil}

-- [[ SETTINGS, PARAMETERS, AUXILIARY DATA ]] --
local mapData -- map data including node connections (edges and edge data), node positions and node radii
local currentRoute
local minPlanCount = 3
local targetWPName

local wpList, manualPath, speedList
local race, noOfLaps

local targetObjectSelectionMode

local edgeDict
------------------------------

-- [[ TRAFFIC ]] --
local currPathIdx = 0
local damageFlag = false
local hornFlag = false
local trafficTable = {}
local trafficBlock = {timer = 0, coef = 0, limit = 6, horn = 0}
local trafficSide = {timer = 0, cTimer = 0, side = 1, timerRange = 6}
local trafficSignal = {hTimer = 0, hLimit = 1}
local intersection = {stopTime = 0, timer = 0, turn = 0}
local avoidCars = 'on'
M.avoidCarsMaster = 'auto'
local changePlanTimer = 0

-----------------------

-- [[ HEAVY DEBUG MODE ]] --
local speedRecordings = {}
local trajecRec = {last = 0}
local routeRec = {last = 0}
local labelRenderDistance = 10

---------------------------------------------------------------------------------------------------------
--Added For Derby -----------------------------------------------------------------------------------------------------
--local crashPos 
--local crashGear
local crashObjID = -1
local bounds = nil
local aiGear = 0
local targetstatus
---------------------------
--End
-----------------------------------------------------------------------------------------------------------

local function aistatus(status, category)
  guihooks.trigger("AIStatusChange", {status=status, category=category})
end

local function getState()
  return M
end

local function stateChanged()
  if playerInfo.anyPlayerSeated then
    guihooks.trigger("AIStateChange", getState())
  end
end

local function setSpeed(speed)
  if type(speed) ~= 'number' then M.routeSpeed = nil else M.routeSpeed = speed end
end

local function setSpeedMode(speedMode)
  if speedMode == 'set' or speedMode == 'limit' or speedMode == 'legal' or speedMode == 'off' then
    M.speedMode = speedMode
  else
    M.speedMode = nil
  end
end

local function resetSpeedModeAndValue()
  M.speedMode = nil -- maybe this should be 'off'
  M.routeSpeed = nil
end

local function setAggressionInternal(v)
  aggression = v and v or M.extAggression
end

local function setAggressionExternal(v)
  M.extAggression = v or M.extAggression
  setAggressionInternal()
  stateChanged()
end

local function setAggressionMode(aggrmode)
  if aggrmode == 'rubberBand' then
    aggressionMode = 'rubberBand'
  else
    aggressionMode = nil
  end
end

local function resetAggression()
  setAggressionInternal()
end

local function setTargetObjectID(id)
  M.targetObjectID = M.targetObjectID ~= objectId and id or -1
  if M.targetObjectID ~= -1 then targetObjectSelectionMode = 'manual' end
end

local function updatePlayerData()
  if mapmgr.objects[M.targetObjectID] and targetObjectSelectionMode == 'manual' then
    player = mapmgr.objects[M.targetObjectID]
	player.id = M.targetObjectID
 elseif tableSize(mapmgr.objects) == 2 then
    if player ~= nil then
      player = mapmgr.objects[player.id]
    else
      for k, v in pairs(mapmgr.objects) do
        if k ~= objectId then
          M.targetObjectID = k
          player = v
          break
        end
      end
      targetObjectSelectionMode = 'auto'
    end
  else
    if player ~= nil and player.active == true then
      player = mapmgr.objects[player.id]
    else
      for k, v in pairs(mapmgr.objects) do
        if k ~= objectId and v.active == true then
          M.targetObjectID = k
          player = v
          break
        end
      end
      targetObjectSelectionMode = 'targetActive'
    end
  end
  mapmgr.objects[objectId] = mapmgr.objects[objectId] or {pos = aiPos, dirVec = aiDirVec}
end

local function driveCar(steering, throttle, brake, parkingbrake)
  input.event("steering", steering, 1)
  input.event("throttle", throttle, 2)
  input.event("brake", brake, 2)
  input.event("parkingbrake", parkingbrake, 2)

  lastCommand.steering = steering
  lastCommand.throttle = throttle
  lastCommand.brake = brake
  lastCommand.parkingbrake = parkingbrake
end

local function driveToTargetDerby(targetPos, throttle, brake)
  if not targetPos then return end
  local debugDrawer = obj.debugDrawProxy
	--Added for Derby
		local ahead
		local pbrake = 0
		local brakeCoef = 1
		local throttleCoef = 1
		local coef = 1
		--local plID = player.id
		--local v = {}
		local ahead2
		--if plID then
			--v.length = obj:getObjectInitialLength(plID) + 0.2
			--v.width = obj:getObjectInitialWidth(plID)
			-- local posFront = obj:getObjectFrontPosition(plID)
			--local pdirVec = player.dirVec
			--v.posFront = pdirVec * 0.2 + posFront
			--v.posRear = pdirVec * (-v.length) + posFront
			--v.paiSpeed = player.vel:length() 
			--ahead2 =  player.pos + player.vel * dt
		--else
			--ahead2 = player.pos
		--end
		if player then
			ahead2 =  player.pos + player.vel * dt
		else	
			ahead2 = targetPos
		end
		--local posFront = obj:getObjectFrontPosition(objectId)
		--posFront = aiDirVec * 0.1 + posFront
		--local plai = min((player.pos - aiPos):length() * 0.02,0.8) 
		ahead = aiDirVec * (-aiLength) + aiPos
		local target3Pos = targetPos
		local targetVec = (ahead - ahead2):normalized()
		local dirTarget = aiDirVec:dot(targetVec)
		targetVec = -targetVec
		dirTarget = -dirTarget
		
		--debugDrawer:drawSphere(0.15, ahead:toFloat3(), color(0,255,0,255))
		--**debugDrawer:drawSphere(0.15, ahead2:toFloat3(), color(0,255,255,255))
		--debugDrawer:drawSphere(0.5, target3Pos:toFloat3(), color(255,255,0,255))
		--local col  = getContrastColor(objectId)	
		--debugDrawer:drawLine(ahead:toFloat3(), ahead2:toFloat3(), col)
		local aheadTarget = aiDirVec:dot(targetVec)
		local dirDiff = -math.asin(aiDirVec:cross(aiUpVec):normalized():dot(targetVec))
		local dirAngle = math.asin(aiRightVec:dot(targetVec))
		local dirVel = aiVel:dot(aiDirVec)
		local absAiSpeed = abs(dirVel)	
		if crash.manoeuvre > 0 then  --reset target
			if not crash.pos and crash.manoeuvre == 4 then
				local crashObj = mapmgr.objects[crashObjID]
					crash.dir = -aiDirVec --crashObj.dirVec
					crash.pos = crashObj.pos
				--debugDrawer:drawText(crash.pos:toFloat3() + float3(0, 0, 2.3), color(0,0,0,255), crash.manoeuvre.."crashManeuver")
			end
			local distFromCrash
			if crash.pos then 
				distFromCrash = aiPos:distance(crash.pos)
			else
				return
			end
				--print(objectId.. " crashManeuver"..crash.manoeuvre.. " Distance" .. distFromCrash)
				--**debugDrawer:drawText(aiPos:toFloat3() + float3(0, 0, .5), color(0,0,0,255), crash.manoeuvre.."Crash  "..distFromCrash)
				--**debugDrawer:drawSphere(0.5, (crash.pos):toFloat3(), color(255,0,0,255))
			if crash.manoeuvre == 5 and distFromCrash < 7 or crash.manoeuvre ~= 5 and distFromCrash > 7 then
				crash.manoeuvre = 0
				throttle = 0
				brake = 0
				--threewayturn.state = 0
				pbrake = 1
				crash.pos = nil
				crash.time = 0
				targetstatus = 2	
				crashObjID = -1
				M.targetObjectID = -1
				driveCar(dirDiff, throttle , brake , pbrake)
				return
			else
				local crashTargetVec = (crash.pos - ahead):normalized()
				local crashTarget = crash.dir:dot(crashTargetVec)		
				dirTarget = -crashTarget				
				dirDiff = -asin(aiDirVec:cross(vec3(obj:getDirectionVectorUp())):dot(crashTargetVec))
				if crash.manoeuvre == 3 then
					throttle = min(max(1 - (aiPos - crash.pos):length() * 0.05, 0.5),1)
				else
					throttle = min(max(0.2 + (aiPos - crash.pos):length() * 0.005, 0.3),0.4)
				end	
				if crash.manoeuvre == 4 then
					crash.time = crash.time + dt
				end
			end
		else
			if absAiSpeed > 0.2 then
			local avgWheelSpeed = 0
			local wheelCount = 0
			local minAbsWheelSpeed = math.huge
			local maxAbsWheelSpeed = 0
			local lwheels = wheels.wheels
			for i = 0, tableSizeC(lwheels) - 1 do
				local wd = lwheels[i]
				if not wd.isBroken then
					wheelCount = wheelCount + wd.isSpeedo
					local absWheelVel = abs(wd.angularVelocity * wd.radius)
					avgWheelSpeed = avgWheelSpeed + absWheelVel * wd.isSpeedo
					if wd.brakeTorque > 0 then
						minAbsWheelSpeed = min(minAbsWheelSpeed, absWheelVel)
					end
					if wd.isPropulsed then
						maxAbsWheelSpeed = max(maxAbsWheelSpeed, absWheelVel)
					end
				end
			end

			-- manual abs and tcs
			if wheelCount > 0 then
				local latVgrad = 0.5 * aiWidth * abs(sensors.gx2) / (aiSpeed + 1e-30)
				avgWheelSpeed = avgWheelSpeed / wheelCount

				-- abs
				local minGradSpeed = absAiSpeed - latVgrad
				if avgWheelSpeed < minGradSpeed and brake > 0 then
					coef = coef * minAbsWheelSpeed / minGradSpeed
				end

				-- tcs
				local maxGradSpeed = max(3, absAiSpeed + latVgrad)
				if maxAbsWheelSpeed > maxGradSpeed and throttle > 0 then
					local redcoef = maxGradSpeed / maxAbsWheelSpeed
					coef = coef * redcoef * redcoef
				end
			end
			end
			if aiSpeed > 1 then
				local rightVel = aiRightVec:dot(aiVel)
				
				local rotVel = min(1, (prevDirVec:projectToOriginPlane(aiUpVec):normalized() - aiDirVec):length() * dt * 10000)
				
				coef = coef * max(0, 1 - abs(rightVel * aiSpeed * 0.05) * min(1, dirAngle * dirAngle * aiSpeed * 6) * rotVel)
				local dirPer = abs(dirDiff) / 1.55
				if dirPer > .70 then
					local temp = 1 - dirPer
					coef = min(max(temp, coef),1)
				end
			end
		end
		aiForceGoFrontTime = max(0, aiForceGoFrontTime - dt)
		if aiCannotMoveTime > 2 and aiForceGoFrontTime == 0 or crash.time > 5 then
			aiCannotMoveTime = 0
			aiForceGoFrontTime = 4
		end
		if aiForceGoFrontTime > 0  then
			--**debugDrawer:drawText(aiPos:toFloat3() + float3(0, 0, .3), color(0,255,0,255), aiForceGoFrontTime.."FrontTime  ")
			--**debugDrawer:drawText(aiPos:toFloat3() + float3(0, 0, .7), color(0,255,0,255), threewayturn.state.."Threewayturn  ")
			--print('aiForceGoFront' ..aiForceGoFrontTime ..' '.. dirTarget)
			if aiForceGoFrontTime == 2 then
				if dirTarget > 0 then
					threewayturn.state = -1
				else
					threewayturn.state = 1
				end
				
			end
			if threewayturn.state > 0 and dirTarget < 0 then
				dirTarget = abs(dirTarget)
			else
				dirTarget =  -dirTarget
			end
			aiCannotMoveTime = 0
			crash.time = 0
		end
			if dirTarget < -.75 or aiSpeed < 3 and dirTarget <= 0 then
				brake = throttle
				throttle = 0 
			elseif dirTarget > .85 or aiSpeed < 3 and dirTarget >= 0 then
				throttle = throttle
				brake = 0 
			else
				if aiSpeed < 1 then
					if dirTarget > 0 then
						throttle = 0.3
						brake = 0
					else
						throttle = 0
						brake = 0.3
					end
				elseif dirTarget > 0 then
					throttle = 0 
					brake = 0
				else
					brake = 0
					throttle = 0
				end
				targetstatus = 2
				--throttle = 0
				--brake = 0
			end 
	
		--print(throttle.. "throtle "..objectId)
		--print(brake.. "brake "..objectId)
		--print(coef.. "coef "..objectId)
		--print(dirDiff.. "dirDiff "..objectId)
		--**debugDrawer:drawText(aiPos:toFloat3() + float3(0, 0, 1.5), color(0,0,0,255), dirTarget.."dirTarget")
		--**debugDrawer:drawText(aiPos:toFloat3() + float3(0, 0, 1.8), color(0,0,255,255), throttle *coef.."throttle")
		--**debugDrawer:drawText(aiPos:toFloat3() + float3(0, 0, 1.3), color(0,0,255,255), brake* coef.."brake")
		driveCar(-dirDiff, throttle * coef, brake * coef, pbrake)
end
local function driveToTarget(targetPos, throttle, brake)	

		local brakeCoef = 1
		local throttleCoef = 1
		local aiPos2 = aiPos
		aiPos = aiDirVec * (-aiLength) + aiPos
		
		local targetVec = (targetPos - aiPos):normalized()

		local dirAngle = math.asin(aiRightVec:dot(targetVec))
		if dirAngle > 0 then
			targetPos = aiRightVec * (-aiWidth * .5) + targetPos
			targetVec = (targetPos - aiPos):normalized()
			dirAngle = math.asin(aiRightVec:dot(targetVec))
		else
			targetPos = aiRightVec * (aiWidth * .5) + targetPos
			targetVec = (targetPos - aiPos):normalized()
			dirAngle = math.asin(aiRightVec:dot(targetVec))
		end
		--debugDrawer:drawSphere(1, targetPos:toFloat3(), color(0,255,0,255))
		--debugDrawer:drawSphere(1, aiPos:toFloat3(), color(0,255,0,255))
		local dirVel = aiVel:dot(aiDirVec)
		local absAiSpeed = abs(dirVel)
		local plan = currentRoute and currentRoute.plan
		
		-- oversteer	
		 if aiSpeed > 1 then
			local rightVel = aiRightVec:dot(aiVel)
			if rightVel * aiRightVec:dot(targetPos - aiPos) > 0 then
				local rotVel = min(1, (prevDirVec:projectToOriginPlane(aiUpVec):normalized() - aiDirVec):length() * dt * 10000)
				throttleCoef = throttleCoef * max(0, 1 - abs(rightVel * aiSpeed * 0.05) * min(1, dirAngle * dirAngle * aiSpeed * 6) * rotVel)
			end
		end

		if plan and plan[3] and dirVel > 3 then
			local p1, p2 = plan[1].pos, plan[2].pos
			local p2p1 = p2 - p1
			local turnRight = p2p1:cross(aiUpVec):normalized()
			local tp2
			if plan.targetSeg and plan.targetSeg > 1 then
				tp2 = (targetPos - p2):normalized():dot(turnRight)
			else
				tp2 = (plan[3].pos - p2):normalized():dot(turnRight)
			end

			local outDeviation = aiDeviationSmoother:value() - aiDeviation * sign(tp2)
			outDeviation = sign(outDeviation) * min(1, abs(outDeviation))
			aiDeviationSmoother:set(outDeviation)
			aiDeviationSmoother:getUncapped(0, dt)
			
	
			if outDeviation > 0 and absAiSpeed > 3 then
				local steerCoef = outDeviation * absAiSpeed * absAiSpeed * min(1, dirAngle * dirAngle * 4)
				local understeerCoef = max(0, steerCoef) * min(1, abs(aiVel:dot(p2p1:normalized()) * 3))
				local noUndersteerCoef = max(0, 1 - understeerCoef)
				throttleCoef = throttleCoef * noUndersteerCoef
				brakeCoef = min(brakeCoef, max(0, 1 - understeerCoef * understeerCoef))
			end
		else
			aiDeviationSmoother:set(0)
		end

		-- wheel speed
		if absAiSpeed > 0.2 then
			local avgWheelSpeed = 0
			local wheelCount = 0
			local minAbsWheelSpeed = math.huge
			local maxAbsWheelSpeed = 0
			local lwheels = wheels.wheels
			for i = 0, tableSizeC(lwheels) - 1 do
				local wd = lwheels[i]
				if not wd.isBroken then
					wheelCount = wheelCount + wd.isSpeedo
					local absWheelVel = abs(wd.angularVelocity * wd.radius)
					avgWheelSpeed = avgWheelSpeed + absWheelVel * wd.isSpeedo
					if wd.brakeTorque > 0 then
						minAbsWheelSpeed = min(minAbsWheelSpeed, absWheelVel)
					end
					if wd.isPropulsed then
						maxAbsWheelSpeed = max(maxAbsWheelSpeed, absWheelVel)
					end
				end
			end

			-- manual abs and tcs
			if wheelCount > 0 then
				local latVgrad = 0.5 * aiWidth * abs(sensors.gx2) / (aiSpeed + 1e-30)
				avgWheelSpeed = avgWheelSpeed / wheelCount

				-- abs
				local minGradSpeed = absAiSpeed - latVgrad
				if avgWheelSpeed < minGradSpeed and brake > 0 then
					brakeCoef = brakeCoef * minAbsWheelSpeed / minGradSpeed
				end

				-- tcs
				
					local maxGradSpeed = max(3, absAiSpeed + latVgrad)
					if maxAbsWheelSpeed > maxGradSpeed and throttle > 0 then
						local redcoef = maxGradSpeed / maxAbsWheelSpeed
						throttleCoef = throttleCoef * redcoef * redcoef
					end
			end
		end
			
		local dirTarget = aiDirVec:dot(targetVec)
		
		local dirDiff = math.asin(aiDirVec:cross(aiUpVec):normalized():dot(targetVec))
	
		if crash.manoeuvre == 1 and dirTarget < aiDirVec:dot(crash.dir) then
			--print('crashManeuver')
			--print(objectId)
			
			driveCar(-fsign(dirDiff), brake * brakeCoef, throttle * throttleCoef, 0)
			return
		else
			crash.manoeuvre = 0
		end

		aiForceGoFrontTime = max(0, aiForceGoFrontTime - dt)
		if threewayturn.state == 1 and aiCannotMoveTime > 1 and aiForceGoFrontTime == 0 then
			threewayturn.state = 0
			aiCannotMoveTime = 0
			aiForceGoFrontTime = 2
		end

		if aiForceGoFrontTime > 0 and dirTarget < 0 then
			dirTarget = -dirTarget
			dirDiff = -dirDiff
		end
		if dirTarget < 0 and currentRoute then
			 local n1, n2, n3 = plan[1], plan[2], plan[3]
			local edgeDist = min((n2 or n1).radiusOrig, n1.radiusOrig) - aiPos:z0():distanceToLine((n3 or n2).posOrig:z0(), n2.posOrig:z0())
			if edgeDist > aiWidth and threewayturn.state == 0 then
				driveCar(fsign(dirDiff), 0.5 * throttleCoef, 0, min(max(aiSpeed - 3, 0), 1))
				
				--print('forward'.. dirTarget)
			else
				  if threewayturn.state == 0 then
					threewayturn.state = 1
					threewayturn.speedDifInt = 0
				end
				local angleModulation = min(max(0, -(dirTarget-0.15)), 1)
				local speedDif = (10 * aggression * angleModulation) - aiSpeed
					threewayturn.speedDifInt = threewayturn.speedDifInt + speedDif * dt
					local pbrake = clamp(sign2(aiDirVec:dot(gravityDir) - 0.17), 0, 1) -- apply parking brake if reversing on an incline >~ 10 deg
				driveCar(-sign2(dirDiff), 0, clamp(0.05 * speedDif + 0.01 * threewayturn.speedDifInt, 0, 1), pbrake)
				--driveCar(-fsign(dirDiff), 0, max(min(abs(dirTarget), 1), 0.4), 0)
			end
		else
			threewayturn.state = 0
			local pbrake =0
			if dirTarget > 0 then
			
				local dirPer = abs(dirDiff) / 1.55
				--local temp = 1 - dirPer
				--throttle = max(min(temp, throttle),0.0)
				--local dirPer = abs(dirDiff) / 1.55
				--if dirPer > 0 and aiSpeed > 1 then
					local temp = 1 - dirPer
					if temp > .95 then
						throttle = throttle * temp
					end
				if aiSpeed > 14 then
					if targetPos.z - aiPos.z > 6 then
						throttle = 0 					
					end
				end
				--end
			
				--if aiSpeed > 1 then
					--if lastCommand.throttle < .2 and throttleCoef < .7 then
							--throttleCoef = min(1, throttleCoef + throttleCoef * dt)
					--end
					
				--if aiSpeed < 1 then
					--throttleCoef = 0.3
				--end
			end
				
				--local testth = throttle * throttleCoef
				--local testbr = brake * brakeCoef
			--print('throttle'..testth..' elec throttle'..electrics.values.throttle)
			--debugDrawer:drawText(aiPos2:toFloat3() + float3(0, 0, 2), color(0,0,0,255), testth.."throttle  ideal"..electrics.values.throttle)
			--debugDrawer:drawText(aiPos2:toFloat3() + float3(0, 0, 2.5), color(0,0,0,255), testbr.."brake  ideal"..electrics.values.brake)
			--print('elec throttle'..electrics.values.throttle)
			--print('thrrotleco'..throttleCoef)
			--print('brake'..testbr..' elec brake'..electrics.values.brake)
			--print('elec brake'..electrics.values.brake)
			--print('brakeco'..brakeCoef)
			--print('Speed'..aiSpeed)
			--print('tspeed'..targetSpeed)
	
			
			--if aiVel:dot(aiDirVec) * max(aiVel:squaredLength() - 1e-2, 0) < 0 then
				--if aiSpeed < 0.15 and targetSpeed <= 1e-5 then
					--pbrake = 0
				--else
					--pbrake = 0
				--end
				--print('normal?'..pbrake)
				--driveCar(dirDiff, 0.4 * throttleCoef, brake, pbrake)
			--else
				--if (aiSpeed > 4 and aiSpeed < 30 and abs(dirDiff) > 0.97 and brake == 0) or (aiSpeed < 0.15 and targetSpeed and targetSpeed <= 1e-5) then
					--pbrake = 0
				--else
					--pbrake = 0
				--end
				--print('normal2?'..pbrake)
				 
		driveCar(dirDiff, throttle * throttleCoef,0,pbrake) -- brake * brakeCoef, pbrake)
				--driveCar(dirDiff, 0.1, 0, 0)
			--end
	end
end
local function aiPosOnPlan(plan, planCount)
  local aiSeg = 1
  local aiXnormOnSeg = 0
  for i = 1, planCount-1 do
    local p0Pos, p1Pos = plan[i].pos, plan[i+1].pos
    local xnorm = aiPos:xnormOnLine(p0Pos, p1Pos)
    if xnorm < 1 then
      if i < planCount - 2 then
        local nextXnorm = aiPos:xnormOnLine(p1Pos, plan[i+2].pos)
        if nextXnorm >= 0 then
          local p1Radius = plan[i+1].radius
          if aiPos:squaredDistance(linePointFromXnorm(p1Pos, plan[i+2].pos, nextXnorm)) <
          square(aiWidth + lerp(p1Radius, plan[i+2].radius, min(1, nextXnorm))) then
            aiXnormOnSeg = nextXnorm
            aiSeg = i + 1
            break
          end
        end
      end
      aiXnormOnSeg = xnorm
      aiSeg = i
      break
    end
  end

  for _ = 1, aiSeg-1 do
    tableRemove(plan, 1)
    planCount = planCount-1
  end

  return aiXnormOnSeg, planCount
end

local function calculateTarget(plan, planCount)
  planCount = planCount or #plan
  local aiXnormOnSeg
  aiXnormOnSeg, planCount = aiPosOnPlan(plan, planCount)
 -- local targetLength = max(aiSpeed * 0.65, 4.5)
	--local targetLength = math.random(6, 15)
  local targetLength = max(max(6.5, aiSpeed * .95), 12)--targetLength = max(aiSpeed * 0.65, 6.5)
  if planCount >= 3 then
    local xnorm = clamp(aiXnormOnSeg, 0, 1)
    local p2Pos = plan[2].pos
    targetLength = max(targetLength, plan[1].pos:distance(p2Pos) * (1-xnorm), p2Pos:distance(plan[3].pos) * xnorm)
  end

  local remainder = targetLength

  local targetPos = plan[planCount].pos
  local targetSeg = max(1, planCount-1)
  local prevPos = linePointFromXnorm(plan[1].pos, plan[2].pos, aiXnormOnSeg) -- aiPos

  for i = 2, planCount do
    local pos = plan[i].pos
    local segVec = pos - prevPos
    local segLen = segVec:length()

    if remainder <= segLen then
      targetSeg = i - 1
      targetPos = prevPos + segVec * remainder / (segLen + 1e-25)

      -- smooth target
      local xnorm = clamp(targetPos:xnormOnLine(prevPos, pos), 0, 1)
      local lp_n1n2 = linePointFromXnorm(prevPos, pos, xnorm * 0.5 + 0.25)
      if xnorm <= 0.5 then
        if i >= 3 then
          targetPos = linePointFromXnorm(linePointFromXnorm(plan[i-2].pos, prevPos, xnorm * 0.5 + 0.75), lp_n1n2, xnorm + 0.5)
        end
      else
        if i <= planCount - 2 then
          targetPos = linePointFromXnorm(lp_n1n2, linePointFromXnorm(pos, plan[i+1].pos, xnorm * 0.5 - 0.25), xnorm - 0.5)
        end
      end
      break
    end
    prevPos = pos
    remainder = remainder - segLen
  end

  return targetPos, targetSeg, aiXnormOnSeg, planCount
end

local function targetsCompatible(baseRoute, newRoute)
  local baseTvec = baseRoute.plan.targetPos - aiPos
  local newTvec = newRoute.plan.targetPos - aiPos
  if aiSpeed < 2 then return true end
  if newTvec:dot(aiDirVec) * baseTvec:dot(aiDirVec) <= 0 then return false end
  local baseTargetRight = baseTvec:cross(aiUpVec):normalized()
  return abs(newTvec:normalized():dot(baseTargetRight)) * aiSpeed < 2
end

local function getMinPlanLen(limLow, speed, accelg)
  -- given current speed, distance required to come to a stop if I can decelerate at 0.2g
  limLow = limLow or 150
  speed = speed or aiSpeed
  accelg = max(0.2, accelg or 0.2)
  return min(550, max(limLow, 0.5 * speed * speed / (accelg * g)))
end

local function pickAiWp(wp1, wp2, dirVec)
  dirVec = dirVec or aiDirVec
  local vec1 = mapData.positions[wp1] - aiPos
  local vec2 = mapData.positions[wp2] - aiPos
  local dot1 = vec1:dot(dirVec)
  local dot2 = vec2:dot(dirVec)
  if (dot1 * dot2) <= 0 then
    if dot1 < 0 then
      return wp2
    end
  else
    if vec2:squaredLength() < vec1:squaredLength() then
      return wp2
    end
  end
  return wp1
end

local function pathExtend(path, newPath)
  if newPath == nil then return end
  local pathCount = #path
  if path[pathCount] ~= newPath[1] then return end
  pathCount = pathCount - 1
  for i = 2, #newPath do
    path[pathCount+i] = newPath[i]
  end
end

-- http://cnx.org/contents/--TzKjCB@8/Projectile-motion-on-an-inclin
local function projectileSqSpeedToRangeRatio(pos1, pos2, pos3)
  local sinTheta = (pos2.z - pos1.z) / pos1:distance(pos2)
  local sinAlpha = (pos3.z - pos2.z) / pos2:distance(pos3)
  local cosAlphaSquared = max(1 - sinAlpha * sinAlpha, 0)
  local cosTheta = sqrt(max(1 - sinTheta * sinTheta, 0)) -- in the interval theta = {-pi/2, pi/2} cosTheta is always positive
  return 0.5 * g * cosAlphaSquared / max(cosTheta * (sinTheta*sqrt(cosAlphaSquared) - cosTheta*sinAlpha), 0)
end

local function inCurvature(vec1, vec2)
  local vec1Sqlen, vec2Sqlen = vec1:squaredLength(), vec2:squaredLength()
  local dot12 = vec1:dot(vec2)
  local cos8sq = min(1, dot12 * dot12 / max(1e-30, vec1Sqlen * vec2Sqlen))

  if dot12 < 0 then
    local minDsq = min(vec1Sqlen, vec2Sqlen)
    local maxDsq = minDsq / max(1e-30, cos8sq)
    if max(vec1Sqlen, vec2Sqlen) > (minDsq + maxDsq) * 0.5 then
      if vec1Sqlen > vec2Sqlen then
        vec1, vec2 = vec2, vec1
        vec1Sqlen, vec2Sqlen = vec2Sqlen, vec1Sqlen
      end
      vec2 = sqrt(0.5 * (minDsq + maxDsq) / max(1e-30, vec2Sqlen)) * vec2
    end
  end

  return 2 * sqrt((1 - cos8sq) / max(1e-30, (vec1 + vec2):squaredLength()))
end

local function getPathLen(path, startIdx, stopIdx)
  if not path then return end
  startIdx = startIdx or 1
  stopIdx = stopIdx or #path
  local positions = mapData.positions
  local pathLen = 0
  for i = startIdx+1, stopIdx do
    pathLen = pathLen + (positions[path[i]] - positions[path[i-1]]):length()
  end

  return pathLen
end

local function waypointInPath(path, waypoint, startIdx, stopIdx)
  if not path or not waypoint then return end
  startIdx = startIdx or 1
  stopIdx = stopIdx or #path
  for i = startIdx, stopIdx do
    if path[i] == waypoint then
      return i
    end
  end
end

local function getPlanLen(plan, planCount)
  planCount = planCount or #plan
  local planLen = 0
  for i = 2, planCount do
    planLen = planLen + (plan[i].pos - plan[i-1].pos):length()
  end
  return planLen, planCount
end

local function buildNextRoute(plan, planCount, path)
  local nextPathIdx = (plan[planCount].pathidx or 0) + 1 -- if the first plan node is the aiPos it does not have a pathIdx value yet

  if race == true and noOfLaps and noOfLaps > 1 and not path[nextPathIdx] then -- in case the path loops
    local loopPathId
    local pathCount = #path
    local lastWayPoint = path[pathCount]
    for i = 1, pathCount do
      if lastWayPoint == path[i] then
        loopPathId = i
        break
      end
    end
    nextPathIdx = 1 + loopPathId -- nextPathIdx % #path
    noOfLaps = noOfLaps - 1
  end

  local nextNodeName = path[nextPathIdx]
  if not nextNodeName then return end

  local n1Pos, oneWay
  local n2 = mapData.graph[nextNodeName]
  if not n2 then return end
  local n2Pos = mapData.positions[nextNodeName]
  local n2Radius = mapData.radius[nextNodeName]
  local validOneWayLink = true -- test for if oneWay road does not merge into a not oneWay road

  if path[nextPathIdx-1] then
    n1Pos = mapData.positions[path[nextPathIdx-1]]
    local link = mapData.graph[path[nextPathIdx-1]][path[nextPathIdx]]
    local legalSpeed
    if link then
      oneWay = link[3]
      if oneWay then
        for k, v in pairs(mapData.graph[path[nextPathIdx]]) do
          if k ~= path[nextPathIdx - 1] and not v[3] and ((n2Pos - n1Pos):normalized()):dot((mapData.positions[k] - n2Pos):normalized()) > 0.2 then
            validOneWayLink = false
          end
        end
      end

      if nextPathIdx > 2 then
        legalSpeed = min(link[4], mapData.graph[path[nextPathIdx-1]][path[nextPathIdx-2]][4])
      else
        legalSpeed = link[4]
      end
    end
    plan[planCount].legalSpeed = legalSpeed
  elseif path[nextPathIdx+1] then -- if previous plan node is the aiPos then get oneWay data from path segment ahead of nextPathIdx
    n1Pos = vec3(aiPos)
    local link = mapData.graph[path[nextPathIdx]][path[nextPathIdx+1]]
    oneWay = link and link[3] -- why do we need the link check here?
  end

  if driveInLaneFlag then
    local lane = 1
    plan.lane = plan.lane or 0
    if oneWay and validOneWayLink then
      if plan.lane ~= 0 then
        lane = plan.lane
      else
        if path[2] then
          local curPathIdx = (plan[1] and plan[1].pathidx) and max(2, plan[1].pathidx) or 2
          local p1Pos = mapData.positions[path[curPathIdx-1]]
          lane = sign((mapData.positions[path[curPathIdx]] - p1Pos):z0():cross(gravityDir):dot(p1Pos - aiPos))
          plan.lane = lane
        end
      end
    else
      plan.lane = 0
    end

    local nVec1
    if path[nextPathIdx-1] then
      nVec1 = (n1Pos - n2Pos):z0():cross(gravityDir):normalized()
    else
      nVec1 = vec3()
    end

    local nVec2
    if path[nextPathIdx+1] then
      nVec2 = (n2Pos - mapData.positions[path[nextPathIdx+1]]):z0():cross(gravityDir):normalized()
    else
      nVec2 = vec3()
    end

    local width = max(n2Radius * 0.5, aiWidth * 0.7)
    local displacement = max(0, n2Radius - width) -- provide a bit more space in narrow roads so other vehicles can overtake
    n2Pos = n2Pos + displacement * lane * (1 - nVec1:dot(nVec2) * 0.5) * (nVec1 + nVec2)
    n2Radius = width
  end

  local lastPlanPos = plan[planCount] and plan[planCount].pos or aiPos
  local vec = (lastPlanPos - n2Pos):z0()
  local manSpeed = speedList and speedList[nextNodeName]

  return {pos = vec3(n2Pos), posOrig = vec3(n2Pos), radius = n2Radius, radiusOrig = n2Radius,
          posz0 = n2Pos:z0(), vec = vec, dirVec = vec:normalized(), turnDir = vec3(0,0,0),
          manSpeed = manSpeed, pathidx = nextPathIdx}
end

local function mergePathPrefix(source, dest, srcStart)
  srcStart = srcStart or 1
  local sourceCount = #source
  local dict = table.new(0, sourceCount-(srcStart-1))
  for i = srcStart, sourceCount do
    dict[source[i]] = i
  end

  local destCount = #dest
  for i = destCount, 1, -1 do
    local srci = dict[dest[i]]
    if srci ~= nil then
      local res = table.new(destCount, 0)
      local resi = 1
      for i1 = srcStart, srci - 1 do
        res[resi] = source[i1]
        resi = resi + 1
      end
      for i1 = i, destCount do
        res[resi] = dest[i1]
        resi = resi + 1
      end

      return res, srci
    end
  end

  return dest, 0
end

local function createPlan(route)
  local path = route.path
  local plan = route.plan
  local planCount = #plan
  local newPlan = {}
  if planCount == 0 then
    local pos1 = aiPos
    newPlan[1] = {posOrig = vec3(pos1), pos = vec3(pos1), posz0 = pos1:z0(),
                  vec = 8 * vec3(-aiDirVec), dirVec = vec3(-aiDirVec), turnDir = vec3(0,0,0),
                  radiusOrig = 2, radius = 2}
  else
    newPlan[1] = plan[1]
  end

  local n = plan[2] or buildNextRoute(plan, planCount, path)
  if not n then
    route.plan = newPlan
    return 1
  end
  newPlan[2] = n
  local newPlanCount = 2
  local newPlanLen = 0
  local j = 1
  local i = 2
  local minPlanLen = getPlanLen()
  repeat
    local curDist = newPlan[j].posOrig:distance(newPlan[j+1].posOrig)
    local xSq = square(newPlanLen + curDist)
    if curDist > min(220, (1e-7*xSq + 1e-3)*xSq + 6) then
      local n1 = newPlan[j]
      local n2 = newPlan[j+1]
      local deflection = (turnDeflection(newPlan, j) + turnDeflection(newPlan, j + 1)):projectToOriginPlane((n1.pos - n2.pos):normalized())
      local pos = (n1.pos + n2.pos - deflection)*0.5
      local vec = (n1.pos - pos):z0()
      n2.vec = (pos - n2.pos):z0()
      n2.dirVec = n2.vec:normalized()
      local origDeflection = (turnDeflection(newPlan, j, 'posOrig') + turnDeflection(newPlan, j + 1, 'posOrig')):projectToOriginPlane((n1.posOrig - n2.posOrig):normalized())
      tableInsert(newPlan, j+1, {posOrig = (n1.posOrig + n2.posOrig - origDeflection)*0.5, pos = pos, posz0 = pos:z0(),
                                vec = vec, dirVec = vec:normalized(), turnDir = vec3(0, 0, 0),
                                radiusOrig = (n1.radiusOrig + n2.radiusOrig)*0.5, radius = (n1.radius + n2.radius)*0.5,
                                pathidx = n2.pathidx})
      newPlanCount = newPlanCount + 1
    else
      j = j + 1
      if j >= newPlanCount then
        i = i + 1
        local n = plan[i] or buildNextRoute(newPlan, newPlanCount, path)
        if not n then break end
        newPlanCount = newPlanCount + 1
        newPlan[newPlanCount] = n
      end
      newPlanLen = newPlanLen + curDist
    end
    -- 'and i >= planCount' ensures all previous plan nodes have been processed
    -- and inserted into the newPlan. solves the issue of having one very long edge.
  until newPlanLen > minPlanLen and newPlanCount >= minPlanCount and i >= planCount
  newPlan[1].pathidx = newPlan[2].pathidx
  route.plan = newPlan

  return newPlanCount
end

local function mergePathPrefix(source, dest, srcStart)
  srcStart = srcStart or 1
  local dict = table.new(0, #source)
  for i = srcStart, #source do
    dict[source[i]] = i
  end

  for i = #dest, 1, -1 do
    local srci = dict[dest[i]]
    if srci ~= nil then
      local res = table.new(#dest, 0)
      local resi = 1
      for i1 = srcStart, srci - 1 do
        res[resi] = source[i1]
        resi = resi + 1
      end
      for i1 = i, #dest do
        res[resi] = dest[i1]
        resi = resi + 1
      end

      return res, srci
    end
  end

  return dest, 0
end
local function planAhead(route, baseRoute)
  if route == nil then return end
  if route.path == nil then
    route.path = {}
    for i = 1, #route do
      route.path[i] = route[i]
      route[i] = nil
    end
    route.plan = {}
  end

  local plan = route.plan
  local planCount

  if baseRoute and not plan[1] then
    planCount = 0
    -- merge from base plan
    local bsrPlan = baseRoute.plan
    if bsrPlan[2] then
      local commonPathEnd
      route.path, commonPathEnd = mergePathPrefix(baseRoute.path, route.path, bsrPlan[2].pathidx)
      if commonPathEnd >= 1 then
        local refpathidx = bsrPlan[2].pathidx - 1
        for i = 1, #bsrPlan do
          local n = bsrPlan[i]
          if n.pathidx > commonPathEnd then break end
          plan[i] = {pos = vec3(n.pos), posOrig = vec3(n.posOrig), posz0 = vec3(n.posz0),
                    vec = vec3(n.vec), dirVec = vec3(n.dirVec), turnDir = vec3(n.turnDir),
                    radius = n.radius, radiusOrig = n.radiusOrig, pathidx = max(1, n.pathidx-refpathidx), legalSpeed = n.legalSpeed}
          planCount = i
        end
        if planCount > bsrPlan.targetSeg then
          plan.targetSeg = bsrPlan.targetSeg
          plan.targetPos = vec3(bsrPlan.targetPos)
          plan.aiSeg = bsrPlan.aiSeg
        end
      end
    end
  end

  if not plan[1] then
    plan[1] = {posOrig = vec3(aiPos), pos = vec3(aiPos), posz0 = aiPos:z0(), vec = 8 * vec3(-aiDirVec),
               dirVec = vec3(-aiDirVec), turnDir = vec3(0,0,0), radiusOrig = 2, radius = 2}
    planCount = 1
  end

  local planLen
  planLen, planCount = getPlanLen(plan, planCount)
  local minPlanLen = getMinPlanLen()
  while not plan[minPlanCount] or planLen < minPlanLen do
    local n = buildNextRoute(plan, planCount, route.path)
    if not n then break end
    planCount = planCount + 1
    plan[planCount] = n
    planLen = planLen + (n.pos-plan[planCount-1].pos):length()
  end

  if not plan[2] then return end
  if not plan[1].pathidx then plan[1].pathidx = plan[2].pathidx end

  do
    local segmentSplitDelay = plan.segmentSplitDelay or 0
    local distOnPlan = 0
    for i = 1, planCount-1 do
      local curDist = plan[i].posOrig:squaredDistance(plan[i+1].posOrig)
      local xSq = square(distOnPlan)
      if curDist > square(min(220, (25e-8*xSq + 1e-5)*xSq + 6)) and distOnPlan < 550 then
        if segmentSplitDelay == 0 then
          local n1, n2 = plan[i], plan[i+1]
          local pos = (n1.pos + n2.pos) * 0.5
          local vec = (n1.pos - pos):z0()
          n2.vec = (pos - n2.pos):z0()
          n2.dirVec = n2.vec:normalized()
          local legalSpeed
          if n2.pathidx > 1 then
			  --dump(mapData.graph)
			  --dump(route.path[n2.pathidx])
			 -- dump(route.path[n2.pathidx-1])			  
					legalSpeed = n2.legalSpeed --mapData.graph[route.path[n2.pathidx]][route.path[n2.pathidx-1]][4]
				
          else
            legalSpeed = n2.legalSpeed
          end
          tableInsert(plan, i+1, {posOrig = (n1.posOrig + n2.posOrig)*0.5, pos = pos, posz0 = pos:z0(),
                                  vec = vec, dirVec = vec:normalized(), turnDir = vec3(0, 0, 0),
                                  radiusOrig = (n1.radiusOrig + n2.radiusOrig)*0.5, radius = (n1.radius + n2.radius)*0.5,
                                  pathidx = n2.pathidx, legalSpeed = legalSpeed})
          planCount = planCount + 1
          segmentSplitDelay = min(5, floor(90/aiSpeed))
        else
          segmentSplitDelay = segmentSplitDelay - 1
        end
        break
      end
      distOnPlan = distOnPlan + sqrt(curDist)
    end
    plan.segmentSplitDelay = segmentSplitDelay
  end

  if plan.targetSeg == nil then
    local aiXnormOnSeg
    plan.targetPos, plan.targetSeg, aiXnormOnSeg, planCount = calculateTarget(plan, planCount)
    plan.aiSeg = 1
  end

  for i = 0, planCount do
    if forces[i] then
      forces[i]:set(0,0,0)
    else
      forces[i] = vec3(0,0,0)
    end
  end

  -- calculate spring forces
  for i = 1, planCount-1 do
    local n1 = plan[i]
    local n2 = plan[i+1]
    local v1 = n1.dirVec
    local v2 = -n2.dirVec
    local turnDir = (v1 + v2):normalized()

    local nforce = (1-threewayturn.state) * max(1 + v1:dot(v2), 0) * 2 * turnDir
    forces[i+1] = forces[i+1] - nforce
    forces[i-1] = forces[i-1] - nforce
    forces[i] = forces[i] + nforce + nforce

    n1.turnDir:set(turnDir)
    n1.speed = 0
  end

  -- other vehicle awareness
  plan.trafficMinProjSpeed = math.huge
  plan.trafficBlockCoef = nil
  if avoidCars ~= 'off' then
    table.clear(trafficTable)
    local trafficTableLen = 0
    for plID, v in pairs(mapmgr.objects) do
      if plID ~= objectId and (M.mode ~= 'chase' or plID ~= player.id or chaseData.suspectState == 'stopped') then
        v.length = obj:getObjectInitialLength(plID) + 0.3
        v.width = obj:getObjectInitialWidth(plID)
        v.targetType = (player and plID == player.id) and M.mode
        if v.targetType == 'follow' then
          v.width = v.width * 4
        end
        local posFront = obj:getObjectFrontPosition(plID)
        local dirVec = v.dirVec
        v.posFront = dirVec * 0.3 + posFront
        v.posRear = dirVec * (-v.length) + posFront
        v.lightbar = ((v.states and v.states.lightbar) and v.states.lightbar > 0) and true or false
        table.insert(trafficTable, v)
        trafficTableLen = trafficTableLen + 1
      end
    end

    local trafficMinSpeedSq = math.huge
    local distanceT = 0
    local aiPathVel = aiVel:dot((plan[2].pos-plan[1].pos):normalized())
    local aiPathVelInv = 1 / abs(aiPathVel + 1e-30)
    local minTrafficDir = 1

    for i = 2, planCount-1 do
      local n1, n2 = plan[i], plan[i+1]
      local n1pos, n2pos = n1.pos, n2.pos
      local n1n2 = n2pos - n1pos
      local n1n2len = n1n2:length()
      local nDir = n1n2 / (n1n2len + 1e-30)
      n1.trafficSqVel = math.huge
      local arrivalT = distanceT * aiPathVelInv

      if damageFlag then
        n1.trafficSqVel = 0
      else
        for j = trafficTableLen, 1, -1 do
          local v = trafficTable[j]
          local plPosFront, plPosRear, plWidth = v.posFront, v.posRear, v.width
          local ai2PlVec = plPosFront - aiPos
          local ai2PlDir = ai2PlVec:dot(aiDirVec)
          local ai2PlSqDist = ai2PlVec:squaredLength()
          if ai2PlDir > 0 then
            local velDisp = arrivalT * v.vel
            plPosFront = plPosFront + velDisp
            plPosRear = plPosRear + velDisp
          end
          local extVec = nDir * (max(aiWidth, plWidth) * 0.5)
          local n1ext, n2ext = n1pos - extVec, n2pos + extVec
          local rnorm, vnorm = closestLinePoints(n1ext, n2ext, plPosFront, plPosRear)

          if M.mode == 'traffic' and v.lightbar and ai2PlSqDist <= 6400 then -- lightbar awareness
            forces[i] = forces[i] + aiRightVec * 3
            forces[i + 1] = forces[i + 1] + aiRightVec * 3
            trafficSide.cTimer = max(2.5, trafficSide.cTimer)
            n1.trafficSqVel = clamp((ai2PlSqDist / 15) - 16, 0, 64)
          end

          local minSqDist = math.huge
          if rnorm > 0 and rnorm < 1 and vnorm > 0 and vnorm < 1 then
            minSqDist = 0
          else
            local rlen = n1n2len + plWidth
            local xnorm = plPosFront:xnormOnLine(n1ext, n2ext) * rlen
            if xnorm > 0 and xnorm < rlen then minSqDist = min(minSqDist, (n1ext+nDir*xnorm-plPosFront):squaredLength()) end

            xnorm = plPosRear:xnormOnLine(n1ext, n2ext) * rlen
            if xnorm > 0 and xnorm < rlen then minSqDist = min(minSqDist, (n1ext+nDir*xnorm-plPosRear):squaredLength()) end

            rlen = v.length + aiWidth
            local v1 = n1ext - plPosRear
            local v1dot = v1:dot(v.dirVec)
            if (v1dot > 0 and v1dot < rlen) then minSqDist = min(minSqDist, (v1 - v1dot * v.dirVec):squaredLength()) end

            v1 = n2ext - plPosRear
            v1dot = v1:dot(v.dirVec)
            if (v1dot > 0 and v1dot < rlen) then minSqDist = min(minSqDist, (v1 - v1dot * v.dirVec):squaredLength()) end
          end

          if minSqDist < square((aiWidth + plWidth) * 0.8) then
            local velProjOnSeg = max(0, v.vel:dot(nDir))
            local middlePos = (plPosFront + plPosRear) * 0.5
            local forceCoef = trafficSide.side * 0.5 * max(0, max(aiSpeed - velProjOnSeg, sign(-nDir:dot(v.dirVec)) * trafficSide.cTimer)) / ((1 + minSqDist) * (1 + distanceT * min(0.1, 1 / (2 * max(0, aiPathVel - v.vel:dot(nDir)) + 1e-30))))

            if v.targetType ~= 'follow' then
              forces[i] = forces[i] - (sign(n1.turnDir:dot(middlePos - n1.posOrig)) * forceCoef) * n1.turnDir
              forces[i+1] = forces[i+1] - (sign(n2.turnDir:dot(middlePos - n2.posOrig)) * forceCoef) * n2.turnDir
            end

            if avoidCars == 'on' and M.mode ~= 'flee' and M.mode ~= 'random' then
              if minSqDist < square((aiWidth + plWidth) * 0.51)  then
                -- obj.debugDrawProxy:drawSphere(0.25, v.posFront:toFloat3(), color(0,0,255,255))
                -- obj.debugDrawProxy:drawSphere(0.25, plPosFront:toFloat3(), color(0,0,255,255))
                table.remove(trafficTable, j)
                trafficTableLen = trafficTableLen - 1
                plan.trafficMinProjSpeed = min(plan.trafficMinProjSpeed, velProjOnSeg)

                n1.trafficSqVel = min(n1.trafficSqVel, velProjOnSeg * velProjOnSeg)
                trafficMinSpeedSq = min(trafficMinSpeedSq, v.vel:squaredLength())
                minTrafficDir = min(minTrafficDir, v.dirVec:dot(nDir))
              end

              if i == 2 and minSqDist < square((aiWidth + plWidth) * 0.6) and ai2PlDir > 0 and v.vel:dot(aiRightVec) * ai2PlVec:dot(aiRightVec) < 0 then
                n1.trafficSqVel = max(0, n1.trafficSqVel - (v.vel - v.vel:dot(aiDirVec) * v.vel):length())
              end
            end
          end
        end
      end
      distanceT = distanceT + n1n2len
    end

    if max(trafficMinSpeedSq, aiSpeed*aiSpeed) < 0.25 then
      plan.trafficBlockCoef = clamp((1 - minTrafficDir) * 0.5, 0, 1)
    else
      plan.trafficBlockCoef = 0
    end
    plan[1].trafficSqVel = plan[2].trafficSqVel
  end

  -- spring force integrator
  for i = 2, planCount do
    local n = plan[i]
    local k = n.turnDir:dot(forces[i])
    local nodeDisplVec = n.pos + fsign(k) * min(abs(k), 0.5) * n.turnDir - n.posOrig
    local nodeDisplLen = nodeDisplVec:length()
    local maxDispl = max(0, n.radiusOrig - aiWidth * (0.35 + 0.3 / (1 + trafficSide.cTimer * 0.1))) * min(1, aggression * (1 + trafficSide.cTimer * 0.3))
    local nodeDisplLenLim = clamp(nodeDisplLen, 0, maxDispl)
    --n.radius = max(n.radiusOrig - fsign(nodeDisplVec:dot(n.turnDir)) * nodeDisplLenLim - distFromEdge, distFromEdge)
    n.radius = max(0, n.radiusOrig - nodeDisplLenLim)
    n.pos = n.posOrig + (nodeDisplLenLim / (nodeDisplLen + 1e-30)) * nodeDisplVec
    n.posz0:set(n.pos:z0())
    n.vec:set(plan[i-1].posz0 - n.posz0)
    n.dirVec:set(n.vec:normalized())
  end

  local targetSeg = plan.targetSeg
  -- smoothly distribute error from planline onto the front segments
  if targetSeg ~= nil and planCount > targetSeg and plan.targetPos and threewayturn.state == 0 then
    local dTotal = 0
    local sumLen = {}
    for i = 2, targetSeg - 1  do
      sumLen[i] = dTotal
      dTotal = dTotal + (plan[i+1].pos - plan[i].pos):length()
    end
    dTotal = dTotal + (plan.targetPos - plan[targetSeg].pos):length()

    dTotal = max(1, dTotal)
    local p1, p2 = plan[1].pos, plan[2].pos
    local dispVec = (aiPos - linePointFromXnorm(p1, p2, aiPos:xnormOnLine(p1, p2))) * (0.5*dt)
    aiDeviation = dispVec:dot((p2-p1):cross(aiUpVec):normalized())

    local dispVecRatio = dispVec / dTotal
    for i = targetSeg - 1, 3, -1 do
      plan[i].pos = plan[i].pos + (dTotal - sumLen[i]) * dispVecRatio
      plan[i].posz0 = plan[i].pos:z0()
      plan[i+1].vec = plan[i].posz0 - plan[i+1].posz0
      plan[i+1].dirVec = plan[i+1].vec:normalized()
    end

    plan[1].pos = p1 + dispVec
    plan[1].posz0 = plan[1].pos:z0()
    plan[2].pos = p2 + dispVec
    plan[2].posz0 = plan[2].pos:z0()
    if plan[3] then
      plan[3].vec = plan[2].posz0 - plan[3].posz0
      plan[3].dirVec = plan[3].vec:normalized()
    end
  end

  plan.targetPos, plan.targetSeg, plan.aiXnormOnSeg, planCount = calculateTarget(plan)
  plan.aiSeg = 1
  plan.planCount = planCount

  -- plan speeds
  local totalAccel = min(aggression, staticFrictionCoef) * g

  local rLast = plan[planCount]
  if route.path[rLast.pathidx+1] or (race and noOfLaps and noOfLaps > 1) then
    rLast.speed = rLast.manSpeed or sqrt(2 * 550 * totalAccel) -- shouldn't this be calculated based on the path length remaining?
  else
    rLast.speed = rLast.manSpeed or 0
  end

  -- speed planning
  for i = planCount-1, 1, -1 do
    local n1 = plan[i]
    local n2 = plan[i+1]

    -- inclination calculation
    local v12d = n2.pos - n1.pos
    local dist = v12d:length() + 1e-30
    v12d = v12d / dist
    local Gf = gravityVec:dot(v12d) -- acceleration due to gravity parallel to road segment, positive when downhill
    local Gt = (gravityVec - v12d * Gf):length() / g -- gravity vec normal to road segment

    local n2SpeedSq = square(n2.speed)

    local n0vec = plan[max(1, i-2)].posz0 - n1.posz0
    local n3vec = n1.posz0 - plan[min(planCount, i + 2)].posz0
    local curvature = min(
            min(inCurvature(n1.vec, n2.vec), inCurvature(n0vec, n3vec)),
            min(inCurvature(n0vec, n2.vec), inCurvature(n1.vec, n3vec))) + 1.6e-7

    local turnSpeedSq = totalAccel * Gt / curvature -- available centripetal acceleration * radius

    -- https://physics.stackexchange.com/questions/312569/non-uniform-circular-motion-velocity-optimization
    --local deltaPhi = 2 * asin(0.5 * n2.vec:length() * curvature) -- phi = phi2 - phi1 = 2 * asin(halfcord / radius)
    local n1SpeedSq = turnSpeedSq * sin(min(asin(min(1, n2SpeedSq / turnSpeedSq)) + 2*curvature*dist, pi*0.5))

    n1SpeedSq = min(n1SpeedSq, n1.trafficSqVel or math.huge)
    n1.trafficSqVel = math.huge

    n1.speed = n1.manSpeed or
                (M.speedMode == 'legal' and n1.legalSpeed and min(n1.legalSpeed, sqrt(n1SpeedSq))) or
                (M.speedMode == 'limit' and M.routeSpeed and min(M.routeSpeed, sqrt(n1SpeedSq))) or
                (M.speedMode == 'set' and M.routeSpeed) or
                sqrt(n1SpeedSq)
  end

  plan.targetSpeed = plan[1].speed + max(0, plan.aiXnormOnSeg) * (plan[2].speed - plan[1].speed)

  return plan
end

local function doTrafficActions()
  -- horn
  if not hornFlag then
    if not damageFlag and trafficBlock.timer >= trafficBlock.horn then
      electrics.horn(true)
      hornFlag = true
    end
  else
    if trafficBlock.timer < trafficBlock.horn then
      electrics.horn(false)
      hornFlag = false
    end
  end

  -- hazard lights
  if mapmgr.objects[objectId] and mapmgr.objects[objectId].damage >= 1000 then
    if not damageFlag then
      trafficSignal.hTimer = trafficSignal.hTimer + dt
      if trafficSignal.hTimer > trafficSignal.hLimit then
        electrics.set_warn_signal(1)
        trafficSignal.hTimer = 0
        damageFlag = true
      end
    end
    return
  else
    damageFlag = false
  end

  -- intersections & turn signals
  local path = currentRoute.path
  local plan = currentRoute.plan

  if not intersection.pos then
    for i = plan[1].pathidx, min(#path - 2, plan[1].pathidx + 3) do
      local nid2 = path[i + 1] -- next node

      if tableSize(mapData.graph[nid2]) > 2 then -- intersection
        local nid1 = path[i]
        local n1Pos = mapData.positions[nid1]
        local n1Radius = mapData.radius[nid1]
        local n2Pos = mapData.positions[nid2]
        local n2Radius = mapData.radius[nid2]
        local nDir = (n2Pos - n1Pos):z0():normalized()
        local drivability = mapData.graph[nid2][nid1][2]

        for k, v in pairs(mapData.graph[nid2]) do
          if k ~= nid1 then -- not previous node
            local linkDir = (mapData.positions[k] - n2Pos):z0():normalized()
            local linkTurn = nDir:dot(linkDir)
            local validTurn = (k == path[i + 2] and abs(linkTurn) < 0.8)

            if validTurn then -- junction turn
              intersection.turn = -sign2(nDir:cross(gravityDir):dot(linkDir))
            end
            if validTurn or v[2] - drivability > 0 then -- junction turn or drivability difference
              intersection.pos = n2Pos - (n2Pos - n1Pos):normalized() * max(4, n2Radius) * 1.5 -- stop point
              intersection.stopTime = 2.5 + max(0, n2Radius - n1Radius)
              break
            end
          end
        end
      end
      if intersection.pos then break end
    end
  else
    local bestDist = math.huge
    local bestIdx
    local distSq = aiPos:squaredDistance(intersection.pos)

    if intersection.stopTime > 0 then
      for i = 1, #plan - 1 do
        if plan[i].manSpeed == 0 then plan[i].manSpeed = nil end
        local dist = currentRoute.plan[i].pos:squaredDistance(intersection.pos)
        if dist < bestDist then
          bestDist = dist
          bestIdx = i
        end
      end

      plan[bestIdx].manSpeed = 0 -- plan node stop; maybe make stops a unique plan property?
    end

    if intersection.timer == 0 and distSq <= square(aiSpeed * 2) then -- approaching intersection
      if intersection.turn < 0 and electrics.values.turnsignal >= 0 then
        electrics.toggle_left_signal()
      elseif intersection.turn > 0 and electrics.values.turnsignal <= 0 then
        electrics.toggle_right_signal()
      end
    end

    if (bestIdx and bestIdx <= 2) or (not bestIdx and distSq <= 100) then -- maybe distance shouldn't be arbitrary
      intersection.timer = intersection.timer + dt -- stopped at intersection
    end

    if intersection.timer > intersection.stopTime then
      plan[1].manSpeed, plan[2].manSpeed = nil, nil
      if distSq >= 400 then -- cleared intersection area (including other branch nodes)
        intersection.pos = nil
        intersection.turn = 0
        intersection.stopTime = 0
        intersection.timer = 0
      end
    end
  end
end

local function resetMapAndRoute()
   mapData = nil
  currentRoute = nil
  race = nil
  noOfLaps = nil
  internalState = 'onroad'
  changePlanTimer = 0
  resetAggression()
  --added for derby--
	bounds = nil
	--end--
end

local function getMapEdges(cutOffDrivability, node)
  -- creates a table (edgeDict) with map edges with drivability > cutOffDrivability
  if mapData ~= nil then
    local allSCC = mapData:scc(node) -- An array of dicts containing all strongly connected components reachable from 'node'.
    local maxSccLen = 0
    local sccIdx
    for i, scc in ipairs(allSCC) do
      -- finds the scc with the most nodes
      local sccLen = scc[0] -- position at which the number of nodes in currentSCC is stored
      if sccLen > maxSccLen then
        sccIdx = i
        maxSccLen = sccLen
      end
      scc[0] = nil
    end
    local currentSCC = allSCC[sccIdx]
    local keySet = {}
    local keySetLen = 0
    edgeDict = {}
    for nid, n in pairs(mapData.graph) do
      if currentSCC[nid] or not driveInLaneFlag then
        for lid, data in pairs(n) do
          if (currentSCC[lid] or not driveInLaneFlag) and (data[2] > cutOffDrivability) then
            local inNode = data[3] or nid
            local outNode = inNode == nid and lid or nid
            keySetLen = keySetLen + 1
            keySet[keySetLen] = {inNode, outNode}
            edgeDict[inNode..'\0'..outNode] = 1
            if not data[3] or not driveInLaneFlag then
              edgeDict[outNode..'\0'..inNode] = 1
            end
          end
        end
      end
    end
    if keySetLen == 0 then return end
    local edge = keySet[math.random(keySetLen)]
    return edge[1], edge[2]
  end
end

local function newManualPath()
  local newRoute, n1, n2, dist
  local offRoad = false

  if manualPath then
    if currentRoute and currentRoute.path then
      pathExtend(currentRoute.path, manualPath)
    else
      newRoute = {plan = {}, path = manualPath}
      currentRoute = newRoute
    end
    manualPath = nil
  elseif wpList then
    if currentRoute and currentRoute.path then
      newRoute = {plan = currentRoute.plan, path = currentRoute.path}
    else
      n1, n2, dist = mapmgr.findClosestRoad(aiPos)

      if n1 == nil or n2 == nil then
        gui.message("Could not find a road network, or closest road is too far", 5, "AI debug")
        dump("Could not find a road network, or closest road is too far")
        return
      end

      if dist > 2 * max(mapData.radius[n1], mapData.radius[n2]) then
        offRoad = true
        local vec1 = mapData.positions[n1] - aiPos
        local vec2 = mapData.positions[n2] - aiPos

        if aiDirVec:dot(vec1) > 0 and aiDirVec:dot(vec2) > 0 then
          if vec1:length() > vec2:length() then
            n1, n2 = n2, n1
          end
        elseif aiDirVec:dot(mapData.positions[n2] - mapData.positions[n1]) > 0 then
          n1, n2 = n2, n1
        end
      elseif aiDirVec:dot(mapData.positions[n2] - mapData.positions[n1]) > 0 then
        n1, n2 = n2, n1
      end

      newRoute = {plan = {}, path = {n1}}
    end

    for i = 0, #wpList-1 do
      local wp1 = wpList[i] or newRoute.path[#newRoute.path]
      local wp2 = wpList[i+1]
      local route = mapData:getPath(wp1, wp2, driveInLaneFlag and 1e4 or 1)
      local routeLen = #route
      if routeLen == 0 or (routeLen == 1 and wp2 ~= wp1) then
        gui.message("Path between waypoints '".. wp1 .."' - '".. wp2 .."' Not Found", 7, "AI debug")
        dump("Path between waypoints '".. wp1 .."' - '".. wp2 .."' Not Found")
        return
      end

      for j = 2, routeLen do
        tableInsert(newRoute.path, route[j])
      end
    end

    wpList = nil

    if not offRoad and newRoute.path[3] and newRoute.path[2] == n2 then
      tableRemove(newRoute.path, 1)
    end

    currentRoute = newRoute
  end
end

local function validateUserInput(list)
  validateInput = nop
  list = list or wpList
  if not list then return end
  --local isValid = wpList[1] and true or false
  local isValid = list[1] and true or false
  for i = 1, #list do -- #wpList
    local nodeAlias = mapmgr.nodeAliases[list[i]]
    if nodeAlias then
      if mapData.graph[nodeAlias] then
        list[i] = nodeAlias
      else
        if isValid then
          gui.message("One or more of the waypoints were not found on the map. Check the game console for more info.", 6, "AI debug")
          dump('The waypoints with the following names could not be found on the Map')
          isValid = false
        end
        print(list[i])
      end
    end
  end

  return isValid
end

local function warningAIDisabled(message)
  gui.message(message, 5, "AI debug")
  M.mode = 'disabled'
  M.updateGFX = nop
  resetMapAndRoute()
  stateChanged()
end

local function offRoadFollowControl()
  if not player or not player.pos or not aiPos or not aiSpeed then return 0, 0, 0 end

  local ai2PlVec = player.pos - aiPos
  local ai2PlDist = ai2PlVec:length()
  local ai2PlDirVec = ai2PlVec / (ai2PlDist + 1e-30)
  local plSpeedFromAI = player.vel:dot(ai2PlDirVec)
  ai2PlDist = max(0, ai2PlDist - 12)
  local targetSpeed = sqrt(max(0, plSpeedFromAI*plSpeedFromAI*plSpeedFromAI / (abs(plSpeedFromAI) + 1e-30) + 2 * 9.81 * min(1, staticFrictionCoef) * ai2PlDist))
  local speedDif = targetSpeed - aiSpeed
  local throttle = clamp(speedDif, 0, 1)
  local brake = clamp(-speedDif, 0, 1)

  return throttle, brake, targetSpeed
end
------------------------------------------------------------------------------
--Added for Derby
local function findTarget(point)
	gdcallback.vehInRequest()
	local vehIn = gdcallback.vehIn
	--dump(vehIn)
	if not vehIn or #vehIn <= 1 then return end
	local bestTarget = {}
	local bestDist = 0.0
	local bestVec = -1
	--local foundTarget = nil
	--local numObj = tableSize(mapmgr.objects)
	local pDist = aiPos:distance(point)
	local ptVec = (point - aiPos):normalized()
	local pVec = math.abs(aiDirVec:dot(ptVec))
	local numVehIn = #vehIn
	for _, v in pairs(vehIn) do
		if not mapmgr.objects[v.vid] then return end
		if v.vid ~= objectId then
			local tpos = mapmgr.objects[v.vid].pos
			local targetVec = (tpos - aiPos):normalized()
			local tVec = math.abs(aiDirVec:dot(targetVec))
			local vDist = aiPos:distance(tpos)
			if tVec > bestVec then
				bestVec = tVec
			end
			if vDist > bestDist then
				bestDist = vDist
			end
			if vDist > pDist and tVec > pVec then
				table.insert(bestTarget, {vid = v.vid, vDist = vDist, tVec = tVec})
			elseif vDist > 15 and tVec > pVec then
				table.insert(bestTarget, {vid = v.vid, vDist = vDist, tVec = tVec})
			elseif vDist > 25 and tVec > .30 then
				table.insert(bestTarget, {vid = v.vid, vDist = vDist, tVec = tVec})
			end
		end
	end
		--dump(bestTarget)
	if bestTarget[1] then
		local target
		if bestTarget[2] then
			for _,i in ipairs(bestTarget) do
				if i.tVec >= bestVec then
					target = i.vid
					break
				elseif i.vDist >= bestDist then
					target = i.vid
					break
				end
			end
		end
		if target then
			M.targetObjectID = -1
			setTargetObjectID(target)
			updatePlayerData()
			targetstatus = 1
			if crash.manoeuvre == 5 then
				--threewayturn.state = 0
				crash.manoeuvre = 0
			end
		else 
			M.targetObjectID = -1
			target = bestTarget[1].vid
			setTargetObjectID(target)
			updatePlayerData()
			targetstatus = 1
			if crash.manoeuvre == 5 then
				--threewayturn.state = 0
				crash.manoeuvre = 0
			end
		end
	else
		--threewayturn.state = 0 
		targetstatus = 2
		crash.manoeuvre = 5
		crash.pos = vec3(point)
		crash.dir = vec3(-aiDirVec)
		player =  {pos = point, dirVec = vec3(0, 0 ,0), vel = vec3(0, 0, 0)}
		M.targetObjectID = 1
	end
	
	--print(M.targetObjectID)
	
end
local function findPoint()
	--local debugDrawer = obj.debugDrawProxy
	local wp1 = bounds[1].pos
	local wp2 = bounds[2].pos
	local wall = {}
	if wp1 ~= nil and wp2 ~= nil then
		wall[1] = float3(aiPos.x, wp1.y, wp1.z) --wp1
		wall[2] = float3(wp1.x, aiPos.y, wp1.z) --wp1a
		wall[3] = float3(wp2.x,aiPos.y,wp2.z) --wp2a
		wall[4] = float3(aiPos.x,wp2.y,wp2.z) --wp2
				
		local col = getContrastColor(objectId)
		--debugDrawer:drawLine(aiPos:toFloat3(), wall[1], col)
		--debugDrawer:drawLine(aiPos:toFloat3(), wall[2], col)
		--debugDrawer:drawLine(aiPos:toFloat3(), wall[3], col)
		--debugDrawer:drawLine(aiPos:toFloat3(), wall[4], col)
	end
	local wallCount = 0
	local wallPos = 0
	local maxWall = 0.00
	
	local tp = 0
	for k,i in pairs(wall) do
		local walld = aiPos:distance(i)
		
		if walld < 3 then 
			wallCount = wallCount + 1
		end
		if maxWall < walld then
			tp = k
			maxWall = walld
		end
	end
	local point = vec3(0, 0, 0)
	
	if wallCount > 1 and crash.manoeuvre ~= 5 then  --if we are stuck in the corner then go back to the center of the arena
		crash.manoeuvre = 5
		--threewayturn.state = 0 
		aiForceGoFrontTime = 0
		targetstatus = 1
		local xpos = (wp1.x + wp2.x) / 2
		local ypos = (wp1.y + wp2.y) / 2
		point = vec3(xpos, ypos, wp1.z)
		crash.pos = point
		crash.dir = vec3(-aiDirVec)
		player =  {pos = point, dirVec = vec3(0, 0 ,0), vel = vec3(0, 0, 0)}
		M.targetObjectID = 1
	else
		local newtarget
		local targetPoint = {}
		if tp == 1 then
			targetPoint[1] = wp1
			targetPoint[2] = vec3(wp1.x, wp2.y, wp1.z)
			
			--print("wall1")
		elseif tp == 2 then
			targetPoint[1] = vec3(wp1.x, wp2.y, wp1.z)
			targetPoint[2] = wp1
				--print("wall2")
		elseif tp == 3 then
			targetPoint[1] = vec3(wp2.x, wp1.y, wp2.z)
			targetPoint[2] = wp2
				--print("wall3")
		elseif tp == 4 then
			targetPoint[1] = wp2
			targetPoint[2] = vec3(wp2.x, wp1.y, wp2.z)	
				--print("wall4")
		end	

		local p1 = math.random(30,70) * .01
		local midPoint = vec3((wp1.x + wp2.x) / 2,(wp1.y + wp2.y) / 2, aiPos.z)
		local dist = p1 * (midPoint:distance(targetPoint[1]))
		local target = dist * ((targetPoint[1] - midPoint):normalized()) + midPoint
		local dist2 = p1 * (midPoint:distance(targetPoint[2]))
		local target2 = dist2 * ((targetPoint[2] - midPoint):normalized()) + midPoint
		local ran = math.random(0,100)
		local dist3 = (ran *.01) * (target:distance(target2))
		point = vec3(dist3 * ((target2 - target):normalized()) + target)	
		if crash.manoeuvre == 5 then
			point = vec3(player.pos)
		end
		if wallCount == 1 then
			crash.manoeuvre = 5
			aiForceGoFrontTime = 0
			crash.pos = point
			crash.dir = vec3(-aiDirVec)
			player =  {pos = point, dirVec = vec3(0, 0 ,0), vel = vec3(0, 0, 0)}
			M.targetObjectID = 1
		end
		return point 
	end
end


-------------------------------------------------------------------------------------
--RECTANCE LINE INTERSECT  --https://love2d.org/wiki/General_math
-------------------------------------------------------------------------------------
local function findIntersect(l1p1x,l1p1y, l1p2x,l1p2y, l2p1x,l2p1y, l2p2x,l2p2y, seg1, seg2)
	local a1,b1,a2,b2 = l1p2y-l1p1y, l1p1x-l1p2x, l2p2y-l2p1y, l2p1x-l2p2x
	local c1,c2 = a1*l1p1x+b1*l1p1y, a2*l2p1x+b2*l2p1y
	local delta,x,y = a1*b2 - a2*b1
	if delta==0 then return false, "The lines are parallel." end
	x,y = (b2*c1-b1*c2)/delta, (a1*c2-a2*c1)/delta
	if seg1 or seg2 then
		if seg1 and not (min(l1p1x,l1p2x) <= x and x <= max(l1p1x,l1p2x) and min(l1p1y,l1p2y) <= y and y <= max(l1p1y,l1p2y)) or
		   seg2 and not (min(l2p1x,l2p2x) <= x and x <= max(l2p1x,l2p2x) and min(l2p1y,l2p2y) <= y and y <= max(l2p1y,l2p2y)) then
			return false, "The lines don't intersect."
		end
	end
	return x,y
end

local function area(x1, y1, x2, y2, x3, y3)
	return abs((x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2)) / 2.0)
end

local function pointInRect( x1, y1, x2, y2, x3, y3, x4, y4, x, y )
        local A = area(x1, y1, x2, y2, x3, y3) + area(x1, y1, x4, y4, x3, y3)
		local A1 = area(x, y, x1, y1, x2, y2)
		local A2 = area(x, y, x2, y2, x3, y3)
		local A3 = area(x, y, x3, y3, x4, y4)
		local A4 = area(x, y, x1, y1, x4, y4)
		local A5 = math.floor(A1 + A2 + A3 + A4)
		A = math.floor(A)
		if A == A5 then	
			return 1
		else	
			return 0
		end
end

local function checkWalls()
	local wp1 = bounds[1].pos
	local wp2 = bounds[2].pos
	local pir = pointInRect(wp1.x, wp2.y, wp1.x, wp1.y, wp2.x,wp1.y,wp2.x,wp2.y, aiPos.x, aiPos.y)
	if pir == 1 then		
		local ahead = aiPos
		local vel = aiSpeed
		if vel < 3 then 
			vel = 4
		end
		local crashDir
		if aiGear < 0 then   -- driving backwards
			vel = vel + 4
			ahead = -aiDirVec * vel + aiPos
			crashDir = aiDirVec
		elseif aiGear >= 0  then -- driving forward
			ahead = aiDirVec * vel + aiPos
			crashDir = aiDirVec
		end
		if ahead then
			local directions = ahead
			local dx = nil
			local dy = nil
			local lineIntersect = nil
			dx, dy = findIntersect(aiPos.x, aiPos.y, directions.x, directions.y, wp1.x, wp1.y, wp1.x, wp2.y,true,true)
			if dx == false then
				dx, dy = findIntersect(aiPos.x, aiPos.y, directions.x, directions.y, wp2.x, wp2.y, wp2.x, wp1.y,true,true)
			end
			if dx == false then
				dx, dy = findIntersect(aiPos.x, aiPos.y, directions.x, directions.y, wp1.x, wp1.y, wp2.x, wp1.y,true,true)
			end
			if dx == false then
				dx, dy = findIntersect(aiPos.x, aiPos.y, directions.x, directions.y, wp2.x, wp2.y, wp1.x, wp2.y,true,true)
			end
			if dx ~= false then
				lineIntersect = vec3(dx, dy, aiPos.z)
				crash.manoeuvre = 3  --hit something
				--threewayturn.state = 0
				aiForceGoFrontTime = 0
				crash.pos = lineIntersect
				crash.dir = vec3(crashDir)
			end
		end
	else
		print('Out of Bounds'..objectId)
		crash.manoeuvre = 5 
		--threewayturn.state = 0
		targetstatus = 1
		local xpos = (wp1.x + wp2.x) / 2
		local ypos = (wp1.y + wp2.y) / 2
		local point = vec3(xpos, ypos, wp1.z)
		crash.pos = point
		crash.dir = vec3(-aiDirVec)
		player =  {pos = point, dirVec = vec3(0, 0 ,0), vel = vec3(0, 0, 0)}
		M.targetObjectID = 1
	end
end	

local function checkHitTarget()
	local aiCollisions = mapmgr.objects[objectId].objectCollisions
	if aiCollisions ~= nil then
		for k,_ in pairs (aiCollisions) do
			if k ~= objectId then
				if crashObjID ~= k then
					crashObjID = k
					crash.manoeuvre = 4
					--threewayturn.state = 0
					M.targetObjectID = 1
				end
			end
		end
	end
end

--Ended


local function getEdges() 
	local wp1 = 'wp_1'
	local wp2 = 'wp_2'
	local dist1 = 0
	--dump(mapData)
	--dump(mapmgr.nodeAliases)
	if not mapData.positions[wp1] then
		wp1, wp2, dist1 = mapmgr.findClosestRoad(aiPos) --Find the Arena Waypoints
		local savedWP
		local savedWPNum
		local maxDist = math.huge
		for k,pos in pairs(mapData.positions) do
			local wp,last = string.find(k, "wp_")
			if wp ~= nil then
				last = last + 1
				local wpNum = string.sub(k, last)	
				if (wpNum % 2 ~= 0) then
					local dist = aiPos:distance(pos)
					if dist < maxDist then
						maxDist = dist
						savedWPNum = wpNum
						savedWP = k
					end		
				end
			end
		end
		if savedWP then	
			savedWPNum = savedWPNum + 1
			local tempwp1 = mapData.positions[savedWP]
			local tempwp2 = mapData.positions['wp_' .. savedWPNum]
			
			local pir = pointInRect(tempwp1.x, tempwp2.y, tempwp1.x, tempwp1.y, tempwp2.x,tempwp1.y,tempwp2.x,tempwp2.y, aiPos.x, aiPos.y)
			if pir == 1 then
				wp1 = savedWP
				wp2 = 'wp_' .. savedWPNum
			end
		end
	end
	if wp1 == nil or wp2 == nil then
		--print('no path') 
		return
	end
	bounds = {}
	bounds[1] = {wp = wp1, pos = mapData.positions[wp1]}
	bounds[2] = {wp = wp2, pos = mapData.positions[wp2]}
	dump(bounds)
	targetstatus = 2
end

M.updateGFX = nop
local function updateGFX(dtGFX)
	dt = dtGFX
  
  if mapData ~= mapmgr.mapData then
    currentRoute = nil
  end

  mapData = mapmgr.mapData

  if mapData == nil then return end
  
  -- local cgPos = obj:calcCenterOfGravity()
  -- aiPos:set(cgPos)
  -- aiPos.z = obj:getSurfaceHeightBelow(cgPos)
  local tmpPos = obj:getFrontPosition()
  aiPos:set(tmpPos)
  aiPos.z = max(aiPos.z - 1, obj:getSurfaceHeightBelow(tmpPos))
  prevDirVec:set(aiDirVec)
  aiDirVec:set(obj:getDirectionVector())
  aiUpVec:set(obj:getDirectionVectorUp())
  aiVel:set(obj:getVelocity())
  aiSpeed = aiVel:length()
  aiWidth = aiWidth or obj:getInitialWidth()
  aiLength = aiLength or obj:getInitialLength()
  aiRightVec = aiDirVec:cross(aiUpVec):normalized()
  staticFrictionCoef = 0.95 * obj:getStaticFrictionCoef()
  
 if trafficBlock.coef > 0 then
    trafficBlock.timer = trafficBlock.timer + dt * trafficBlock.coef
  else
    trafficBlock.timer = trafficBlock.timer * 0.8
  end

  if max(lastCommand.brake, lastCommand.throttle) > 0.3 and aiSpeed < 1 then
    aiCannotMoveTime = aiCannotMoveTime + dt
  else
    aiCannotMoveTime = 0
  end
  
  if aiSpeed < 3 then
    trafficSide.cTimer = trafficSide.cTimer + dt
    trafficSide.timer = (trafficSide.timer + dt) % (2 * trafficSide.timerRange)
    trafficSide.side = sign2(trafficSide.timerRange - trafficSide.timer)
  else
    trafficSide.cTimer = max(0, trafficSide.cTimer - dt)
    trafficSide.timer = 0
    trafficSide.side = 1
  end

  changePlanTimer = max(0, changePlanTimer - dt)

 ------------------ DERBY MODE -----------------
	if M.mode == 'derby' then
		aiGear = drivetrain.gear
		if not bounds then
			
			validateInput = validateUserInput
			getEdges()
		else
			if crash.manoeuvre == 0 then
				checkHitTarget()
			end
			local point = findPoint()
			if targetstatus == 2 then
				findTarget(point)
			end
			if crash.manoeuvre ~= 5 then
				checkWalls()
			end
			if player then
				if mapmgr.objects[M.targetObjectID] then
					player = mapmgr.objects[M.targetObjectID] --need to update the player pos each time before creating a route
				end
				setAggressionInternal(min(0.6 + 0.001 * (aiPos - player.pos):length(), 1))
				currentRoute = nil
				driveToTargetDerby(player.pos, 1, 0)
				return	
			else
				gui.message("No vehicle to Chase", 5, "AI debug")
				return
			end
		end
		
		if not currentRoute then return end

	------------------ RACE MODE -----------------
	elseif M.mode == 'manual' then
		if validateInput() or targetstatus == 2 then 
			newManualPath() 
			wpList = savewpList
			local savedWP 
			local path1 = currentRoute.path[1]
			local paths = #currentRoute.path
			if race == true and path1 ~= currentRoute.path[paths] then 
				for k,v in ipairs(currentRoute.path) do
					if v == path1 and k ~= 1 then
						savedWP = k
						break
					end
				end
				if savedWP then
					for i = (savedWP + 1), paths do
						table.remove(currentRoute.path,i)
					end
				else
					currentRoute.path[paths + 1]= path1
				end
			end
		end
	
		if aggressionMode == 'rubberBand' then
			updatePlayerData()
			if player ~= nil then
				if (aiPos - player.pos):dot(aiDirVec) > 0 then
					setAggressionInternal(max(min(0.1 + max((150 - (player.pos - aiPos):length())/150, 0), 1), 0.5))
				else
					setAggressionInternal(0.95)
				end
			end
		end

		planAhead(currentRoute)

		if not currentRoute then  return end

  ------------------ STOP MODE ----------------
	elseif M.mode == 'stop' then
		--print('stop')
		if aiVel:dot(aiDirVec) > 0 then
			driveCar(0, 0, 0.5, 0)
		else
			driveCar(0, 1, 0, 0)
		end
		if aiSpeed < 0.08 then --  or aiVel:dot(aiDirVec) < 0
			driveCar(0, 0, 0, 0) -- only parkingbrake
			M.mode = 'disabled'
			M.manualTargetName = nil
			M.updateGFX = nop
			resetMapAndRoute()
			stateChanged()
			return
		end
	end	

	if currentRoute then
		local plan = currentRoute.plan
		local targetPos = plan.targetPos
		local aiSeg = plan.aiSeg
		local targetSpeed = plan.targetSpeed
			trafficBlock.coef = plan.trafficBlockCoef or trafficBlock.coef
		if aiUpVec:dot(gravityVec) > 0 then -- vehicle upside down
			return
		end

	 if not targetPos or ((targetSpeed < 0.05 and aiSpeed < 0.1 or (targetPos - aiPos):dot(aiDirVec) < 0) and aiSeg == #plan-1) then
      if M.mode == 'span' then
        local path = currentRoute.path
        for i = 1, #path - 1 do
          local key = path[i]..'\0'..path[i+1]
          -- in case we have gone over an edge that is not in the edgeDict list
          edgeDict[key] = edgeDict[key] and (edgeDict[key] * 20)
        end
      end
		
		--print('done')
		--dump(currentRoute)
		driveCar(0, 0, 0, 1)
		aistatus('route done', 'route')
		gui.message("Route done", 5, "AI debug")
		currentRoute = nil
		speedRecordings = {}
		return
	end
	
-- Throttle and Brake control
  -- Throttle and Brake control
    local speedDif = targetSpeedSmoother:get(targetSpeed, dt) - aiSpeed
    local lowSpeedDif = (speedDif - clamp((aiSpeed - 2) * 0.5, 0, 1)) * 0.5
    local throttle = clamp(lowSpeedDif, 0, 1)
    local brake = clamp(-speedDif, 0, 1)
	
	-- TODO: this still runs if there is no currentPlan, but raises error if there is no targetSpeed
    if not controller.isFrozen and aiSpeed < 0.1 and targetSpeed > 0.5 and (lastCommand.throttle ~= 0 or lastCommand.brake ~= 0) then
      crash.time = crash.time + dt
      if crash.time > 1 then
        crash.dir = vec3(aiDirVec)
        crash.manoeuvre = 1
      end
    else
      crash.time = 0
    end	
		
-----------------------------------------------------------------------------------------------
  
      driveToTarget(targetPos, throttle , brake)
 end
end

local function debugDraw(focusPos)
  local debugDrawer = obj.debugDrawProxy

  if M.mode == 'script' and scriptai ~= nil then
    scriptai.debugDraw()
  end

  if currentRoute then
    local plan = currentRoute.plan
    local targetPos = plan.targetPos
    local targetSpeed = plan.targetSpeed
    if targetPos then
      local tmpfl3 = float3(targetPos:xyz())
      debugDrawer:drawSphere(0.25, tmpfl3, color(255,0,0,255))

      local aiSeg = plan.aiSeg
      local shadowPos = currentRoute.plan[aiSeg].pos + plan.aiXnormOnSeg * (plan[aiSeg+1].pos - plan[aiSeg].pos)
      tmpfl3:set(shadowPos:xyz())
      local blue = color(0,0,255,255)
      debugDrawer:drawSphere(0.25, tmpfl3, blue)

      for plID, _ in pairs(mapmgr.objects) do
        if plID ~= objectId then
          local plPosFront = vec3(obj:getObjectFrontPosition(plID))
          tmpfl3:set(plPosFront:xyz())
          debugDrawer:drawSphere(0.25, tmpfl3, blue)
        end
      end

      if player then
        tmpfl3:set(player.pos:xyz())
        debugDrawer:drawSphere(0.3, tmpfl3, color(0,255,0,255))
      end
    end

    if M.debugMode == 'target' then
      if mapData and mapData.graph and currentRoute.path then
        local p = mapData.positions[currentRoute.path[#currentRoute.path]]:toFloat3()
        --debugDrawer:drawSphere(4, p, color(255,0,0,100))
        --debugDrawer:drawText(p + float3(0, 0, 4), color(0,0,0,255), 'Destination')
      end

    elseif M.debugMode == 'route' then
      if currentRoute.path then
        local p = mapData.positions[currentRoute.path[#currentRoute.path]]:toFloat3()
        debugDrawer:drawSphere(4, p, color(255,0,0,100))
        debugDrawer:drawText(p + float3(0, 0, 4), color(0,0,0,255), 'Destination')
      end

      local maxLen = 700
      local last = routeRec.last
      local len = min(#routeRec, maxLen)
      if len == 0 or (routeRec[last] - aiPos:toFloat3()):length() > 7 then
        last = 1 + last % maxLen
        routeRec[last] = aiPos:toFloat3()
        len = min(len+1, maxLen)
        routeRec.last = last
      end

      local fl3 = float3(0.7, aiWidth, 0.7)
      local black = color(0,0,0,128)
      for i = 1, len-1 do
        debugDrawer:drawSquarePrism(routeRec[1+(last+i-1)%len], routeRec[1+(last+i)%len], fl3, fl3, black)
      end

      if currentRoute.plan[1].pathidx then
        local positions = mapData.positions
        local path = currentRoute.path
        fl3 = fl3 + float3(0, aiWidth, 0)
        local transparentRed = color(255,0,0,120)
        local tmp1fl3 = float3(0, 0, 0)
        local tmp2fl3 = float3(0, 0, 0)
        for i = currentRoute.plan[1].pathidx, #path - 1 do
          tmp1fl3:set(positions[path[i]]:xyz())
          tmp2fl3:set(positions[path[i+1]]:xyz())
          debugDrawer:drawSquarePrism(tmp1fl3, tmp2fl3, fl3, fl3, transparentRed)
        end
      end

    elseif M.debugMode == 'speeds' then
      if plan[1] then
        local red = color(255,0,0,200) -- getContrastColor(objectId)
        local black = color(0, 0, 0, 255)
        local prevSpeed = -1
        local drawLen = 0
        local prevPoint = plan[1].pos:toFloat3()
        local tmpfl3 = float3(0, 0, 0)
        local p = float3(0, 0, 0)
        for i = 1, #plan do
          local n = plan[i]
          p:set(n.pos:xyz())
          local speed = (n.speed >= 0 and n.speed) or prevSpeed
          tmpfl3:set(0, 0, speed * 0.2)
          local p1 = p + tmpfl3
          debugDrawer:drawCylinder(p, p1, 0.03, red)
          debugDrawer:drawCylinder(prevPoint, p1, 0.05, red)
          debugDrawer:drawText(p1, black, strFormat("%2.0f", speed*3.6).." kph")
          prevPoint = p1
          prevSpeed = speed
          --drawLen = drawLen + n.vec:length()
          -- if traffic and traffic[i] then
          --   for _, data in ipairs(traffic[i]) do
          --     local plPosOnPlan = linePointFromXnorm(n.pos, plan[i+1].pos, data[2])
          --     debugDrawer:drawSphere(0.25, plPosOnPlan:toFloat3(), color(0,255,0,100))
          --   end
          -- end
          -- if drawLen > 150 then break end
        end

        local aiPosFlt = aiPos:toFloat3()
        local speedRecLen = #speedRecordings

        if speedRecLen == 0 or (speedRecordings[speedRecLen][1]-aiPosFlt):length() > 0.25 then
          tableInsert(speedRecordings, {aiPosFlt, aiSpeed, targetSpeed, lastCommand.brake, lastCommand.throttle})
          speedRecLen = speedRecLen + 1
        end

        local lastEntry
        local zOffSet = float3(0,0,0.5)
        local yellow, blue = color(255,255,0,200), color(0,0,255,200)
        local tmp2fl3 = float3(0, 0, 0)
        for i = 1, speedRecLen do
          local v = speedRecordings[i]
          if lastEntry then
            -- actuall speed
            tmpfl3:set(0, 0, lastEntry[2] * 0.2)
            tmp2fl3:set(0, 0, v[2] * 0.2)
            debugDrawer:drawCylinder(lastEntry[1] + tmpfl3, v[1] + tmp2fl3, 0.02, yellow)

            -- target speed
            tmpfl3:set(0, 0, lastEntry[3] * 0.2)
            tmp2fl3:set(0, 0, v[3] * 0.2)
            debugDrawer:drawCylinder(lastEntry[1] + tmpfl3, v[1] + tmp2fl3, 0.02, blue)
          end

          tmpfl3:set(0, 0, v[3] * 0.2)
          debugDrawer:drawCylinder(v[1], v[1] + tmpfl3, 0.01, blue)

          if (focusPos - v[1]):length() < labelRenderDistance then
            tmpfl3:set(0, 0, v[2] * 0.2)
            debugDrawer:drawText(v[1] + tmpfl3 + zOffSet, yellow, strFormat("%2.0f", v[2]*3.6).." kph")

            tmpfl3:set(0, 0, v[3] * 0.2)
            debugDrawer:drawText(v[1] + tmpfl3 + zOffSet, blue, strFormat("%2.0f", v[3]*3.6).." kph")
          end
          lastEntry = v
        end

        if speedRecLen > 175 then
          tableRemove(speedRecordings, 1)
        end
      end

      -- Debug Throttle brake application
      local maxLen = 175
      local len = min(#trajecRec, maxLen)
      local last = trajecRec.last
      local aiPosFlt = aiPos:toFloat3()
      if len == 0 or (trajecRec[last][1]-aiPosFlt):length() > 0.25 then
        last = 1 + last % maxLen
        trajecRec[last] = {aiPosFlt, lastCommand.throttle, lastCommand.brake}
        len = min(len+1, maxLen)
        trajecRec.last = last
      end

      local fl3 = float3(0.7, aiWidth, 0.7)
      for i = 1, len-1 do
        local n = trajecRec[1+(last+i)%len]
        debugDrawer:drawSquarePrism(trajecRec[1+(last+i-1)%len][1], n[1], fl3, fl3, color(255*sqrt(abs(n[3])), 255*sqrt(n[2]), 0, 100))
      end

      -- Player segment visual debug for chase / follow mode
      -- if chaseData.playerRoad then
      --   local col1, col2
      --   if internalState == 'tail' then
      --     col1 = color(0,0,0,200)
      --     col2 = color(0,0,0,200)
      --   else
      --     col1 = color(255,0,0,100)
      --     col2 = color(0,0,255,100)
      --   end
      --   local plwp1 = chaseData.playerRoad[1]
      --   debugDrawer:drawSphere(2, mapData.positions[plwp1]:toFloat3(), col1)
      --   local plwp2 = chaseData.playerRoad[2]
      --   debugDrawer:drawSphere(2, mapData.positions[plwp2]:toFloat3(), col2)
      -- end

    elseif M.debugMode == 'trajectory' then
      -- Debug Curvatures
      -- local plan = currentRoute.plan
      -- if plan ~= nil then
      --   local prevPoint = plan[1].pos:toFloat3()
      --   for i = 1, #plan do
      --     local p = plan[i].pos:toFloat3()
      --     local v = plan[i].curvature or 1e-10
      --     local scaledV = abs(1000 * v)
      --     debugDrawer:drawCylinder(p, p + float3(0, 0, scaledV), 0.06, color(abs(min(fsign(v),0))*255,max(fsign(v),0)*255,0,200))
      --     debugDrawer:drawText(p + float3(0, 0, scaledV), color(0,0,0,255), strFormat("%5.4e", v))
      --     debugDrawer:drawCylinder(prevPoint, p + float3(0, 0, scaledV), 0.06, col)
      --     prevPoint = p + float3(0, 0, scaledV)
      --   end
      -- end

      -- Debug Planned Speeds
      if plan[1] then
        local col = getContrastColor(objectId)
        local prevPoint = plan[1].pos:toFloat3()
        local prevSpeed = -1
        local drawLen = 0
        for i = 1, #plan do
          local n = plan[i]
          local p = n.pos:toFloat3()
          local v = (n.speed >= 0 and n.speed) or prevSpeed
          local p1 = p + float3(0, 0, v*0.2)
          --debugDrawer:drawLine(p + float3(0, 0, v*0.2), (n.pos + n.turnDir):toFloat3() + float3(0, 0, v*0.2), col)
          debugDrawer:drawCylinder(p, p1, 0.03, col)
          debugDrawer:drawCylinder(prevPoint, p1, 0.05, col)
          debugDrawer:drawText(p1, color(0,0,0,255), strFormat("%2.0f", v*3.6) .. " kph")
          prevPoint = p1
          prevSpeed = v
          drawLen = drawLen + n.vec:length()
          if drawLen > 80 then break end
        end
      end

      -- Debug Throttle brake application
      local maxLen = 175
      local len = min(#trajecRec, maxLen)
      local last = trajecRec.last
      if len == 0 or (trajecRec[last][1] - aiPos:toFloat3()):length() > 0.25 then
        last = 1 + last % maxLen
        trajecRec[last] = {aiPos:toFloat3(), lastCommand.throttle, lastCommand.brake}
        len = min(len+1, maxLen)
        trajecRec.last = last
      end

      local fl3 = float3(0.7, aiWidth, 0.7)
      for i = 1, len-1 do
        local n = trajecRec[1+(last+i)%len]
        debugDrawer:drawSquarePrism(trajecRec[1+(last+i-1)%len][1], n[1], fl3, fl3, color(255*sqrt(abs(n[3])), 255*sqrt(n[2]), 0, 100))
      end
    end
  end
end

local function setAvoidCars(v)
  M.avoidCarsMaster = v
  if M.avoidCarsMaster == 'off' or M.avoidCarsMaster == 'on' then
    avoidCars = M.avoidCarsMaster
  else
    avoidCars = M.mode == 'manual' and 'off' or 'on'
  end
  stateChanged()
end

local function driveInLane(v)
  if v == 'on' then
    M.driveInLaneFlag = 'on'
    driveInLaneFlag = true
  else
    M.driveInLaneFlag = 'off'
    driveInLaneFlag = false
  end
  stateChanged()
end

local function setMode(mode)
  if M.avoidCarsMaster == 'off' or M.avoidCarsMaster == 'on' then
    avoidCars = M.avoidCarsMaster
  else
    avoidCars = (mode == 'manual' or (mode == nil and M.mode == 'manual')) and 'off' or 'on'
  end

  if mode ~= nil then
    M.mode = mode
  end

  if M.mode ~= 'script' then
    if M.mode ~= 'disabled' and M.mode ~= 'stop' then
      resetMapAndRoute()

      mapmgr.requestMap() -- a map request is also performed in the startFollowing function of scriptai
      M.updateGFX = updateGFX
      targetSpeedSmoother = newTemporalSmoothingNonLinear(math.huge, 0.2, vec3(obj:getVelocity()):length())

      if controller.mainController then
        controller.mainController.setGearboxMode("arcade")
      end
    end

    if M.mode == 'disabled' then
      driveCar(0, 0, 0, 0)
      M.updateGFX = nop
      currentRoute = nil
    end

    if M.mode == 'traffic' then
      setSpeedMode('legal')
      obj:setSelfCollisionMode(2)
      obj:setAerodynamicsMode(2)
    else
      obj:setSelfCollisionMode(1)
      obj:setAerodynamicsMode(1)
    end

    stateChanged()
  end

  speedRecordings = {}
  trajecRec = {last = 0}
  routeRec = {last = 0}
end

local function reset() -- called when the user pressed I
  M.manualTargetName = nil
  trafficBlock = {timer = 0, coef = 0, limit = 6, horn = 0}

  trafficSide.timer = 0
  trafficSide.cTimer = 0
  trafficSide.side = 1

  trafficSignal = {hTimer = 0, hLimit = 1}
  intersection = {stopTime = 0, timer = 0, turn = 0}
  if M.mode ~= 'disabled' then
    driveCar(0, 0, 0, 0)
  end
  setMode() -- some scenarios don't work if this is changed to setMode('disabled')
  stateChanged()
end

local function resetLearning()
end

local function setVehicleDebugMode(newMode)
  tableMerge(M, newMode)
  if M.debugMode ~= 'trajectory' then
    trajecRec = {last = 0}
  end
  if M.debugMode ~= 'route' then
    routeRec = {last = 0}
  end
  if M.debugMode ~= 'speeds' then
    speedRecordings = {}
  end
  if M.debugMode ~= 'off' then
    M.debugDraw = debugDraw
  else
    M.debugDraw = nop
  end
end

local function setState(newState)
  tableMerge(M, newState)
  setAggressionExternal(M.extAggression)
  setMode()
  setVehicleDebugMode(M)
  setTargetObjectID(M.targetObjectID)
end

local function setTarget(wp)
  M.manualTargetName = wp
  validateInput = validateUserInput
  wpList = {wp}
end

local function setPath(path)
  manualPath = path
  validateInput = validateUserInput
end

local function driveUsingPath(arg)
  --[[ At least one argument of either path or wpTargetList must be specified. All other arguments are optional.

  * path: A sequence of waypoint names that form a path by themselves to be followed in the order provided.
  * wpTargetList: Type: A sequence of waypoint names to be used as succesive targets ex. wpTargetList = {'wp1', 'wp2'}.
                  Between any two consequitive waypoints a shortest path route will be followed.

  -- Optional Arguments --
  * wpSpeeds: Type: (key/value pairs, key: "node_name", value: speed, number in m/s)
              Define target speeds for individual waypoints. The ai will try to meet this speed when at the given waypoint.
  * noOfLaps: Type: number. Default value: nil
              The number of laps if the path is a loop. If not defined, the ai will just follow the succesion of waypoints once.
  * routeSpeed: A speed in m/s. To be used in tandem with "routeSpeedMode".
                Type: number
  * routeSpeedMode: Values: 'limit': the ai will not go above the 'routeSpeed' defined by routeSpeed.
                            'set': the ai will try to always go at the speed defined by "routeSpeed".
  * driveInLane: Values: 'on' (anything else is considered off/inactive)
                 When 'on' the ai will keep on the correct side of the road on two way roads.
                 This also affects pathFinding in that when this option is active ai paths will traverse roads in the legal direction if posibble.
                 Default: inactive
  * aggression: Value: 0.3 - 1. The aggression value with which the ai will drive the route.
                At 1 the ai will drive at the limit of traction. A value of 0.3 would be considered normal every day driving, going shopping etc.
                Default: 0.3
  * avoidCars: Values: 'on' / 'off'.  When 'on' the ai will be aware of (avoid crashing into) other vehicles on the map. Default is 'off'
  * examples:
  ai.driveUsingPath{ wpTargetList = {'wp1', 'wp10'}, driveInLane = 'on', avoidCars = 'on', routeSpeed = 35, routeSpeedMode = 'limit', wpSpeeds = {wp1 = 10, wp2 = 40}, aggression = 0.3}
  In the above example the speeds set for wp1 and wp2 will take precedence over "routeSpeed" for the specified nodes.
  --]]

  if (arg.wpTargetList == nil and arg.path == nil and arg.script == nil) or
    (type(arg.wpTargetList) ~= 'table' and type(arg.path) ~= 'table' and type(arg.script) ~= 'table') or
    (arg.wpSpeeds ~= nil and type(arg.wpSpeeds) ~= 'table') or
    (arg.noOfLaps ~= nil and type(arg.noOfLaps) ~= 'number') or
    (arg.routeSpeed ~= nil and type(arg.routeSpeed) ~= 'number') or
    (arg.routeSpeedMode ~= nil and type(arg.routeSpeedMode) ~= 'string') or
    (arg.driveInLane ~= nil and type(arg.driveInLane) ~= 'string') or
    (arg.aggression ~= nil and type(arg.aggression) ~= 'number')
  then
    return
  end

  if arg.resetLearning then
    resetLearning()
  end

  setState({mode = 'manual'})

  noOfLaps = arg.noOfLaps and max(arg.noOfLaps, 1) or 1
  wpList = arg.wpTargetList
  manualPath = arg.path
  validateInput = validateUserInput
  avoidCars = arg.avoidCars or 'off'

  if noOfLaps > 1 and wpList[2] and wpList[1] == wpList[#wpList] then
    race = true
  end

  speedList = arg.wpSpeeds or {}
  setSpeed(arg.routeSpeed)
  setSpeedMode(arg.routeSpeedMode)

  driveInLane(arg.driveInLane)

  setAggressionExternal(arg.aggression)
  stateChanged()
end

local function spanMap(cutOffDrivability)
  M.cutOffDrivability = cutOffDrivability or 0
  setState({mode = 'span'})
  stateChanged()
end

local function setCutOffDrivability(drivability)
  M.cutOffDrivability = drivability or 0
  stateChanged()
end

local function onDeserialized(v)
  setState(v)
  stateChanged()
end

local function dumpCurrentRoute()
  dump(currentRoute)
end

local function startRecording()
  M.mode = 'script'
  scriptai = require("scriptai")
  scriptai.startRecording()
  M.updateGFX = scriptai.updateGFX
end

local function stopRecording()
  M.mode = 'disabled'
  scriptai = require("scriptai")
  local script = scriptai.stopRecording()
  M.updateGFX = scriptai.updateGFX
  return script
end

local function startFollowing(...)
  M.mode = 'script'
  scriptai = require("scriptai")
  scriptai.startFollowing(...)
  M.updateGFX = scriptai.updateGFX
end

local function scriptStop()
  M.mode = 'disabled'
  scriptai = require("scriptai")
  scriptai.scriptStop()
  M.updateGFX = scriptai.updateGFX
end

local function scriptState()
  scriptai = require("scriptai")
  return scriptai.scriptState()
end

local function setScriptDebugMode(mode)
  scriptai = require("scriptai")
  if mode == nil or mode == 'off' then
    M.debugMode = 'all'
    M.debugDraw = nop
    return
  end

  M.debugDraw = debugDraw
  scriptai.debugMode = mode
end

local function isDriving()
  return M.updateGFX == updateGFX or (scriptai ~= nil and scriptai.isDriving())
end

-- public interface
M.driveInLane = driveInLane
M.stateChanged = stateChanged
M.reset = reset
M.setMode = setMode
M.setAvoidCars = setAvoidCars
M.setTarget = setTarget
M.setPath = setPath
M.setSpeed = setSpeed
M.setSpeedMode = setSpeedMode
M.setVehicleDebugMode = setVehicleDebugMode
M.setState = setState
M.getState = getState
M.debugDraw = nop
M.driveUsingPath = driveUsingPath
M.setAggressionMode = setAggressionMode
M.setAggression = setAggressionExternal
M.onDeserialized = onDeserialized
M.setTargetObjectID = setTargetObjectID
M.dumpCurrentRoute = dumpCurrentRoute
M.spanMap = spanMap
M.setCutOffDrivability = setCutOffDrivability
M.resetLearning = resetLearning
M.isDriving = isDriving

-- scriptai
M.startRecording = startRecording
M.stopRecording = stopRecording
M.startFollowing = startFollowing
M.stopFollowing = scriptStop
M.scriptStop = scriptStop
M.scriptState = scriptState
M.setScriptDebugMode = setScriptDebugMode
return M
