
var app = angular.module('derbyApp', ['ngMaterial','ui.router']);

app.run(function() {
  console.log('Derby App Started');
});

app.factory('$localstorage', ['$window', function($window) {
  return {
	saveObject: function(key,value) {
		bngApi.engineLua('gdcallback.saveObject(' + bngApi.serializeToLua(key) + ',' + bngApi.serializeToLua(value) + ')');
	},
	saveTable: function(table) {
		bngApi.engineLua('gdcallback.saveTable(' + bngApi.serializeToLua(table) + ')');
	},
	loadObject: function(key, defaultValue,saved) {
		if(saved !== undefined) {
			if(saved !== null || saved !== undefined ){
				if (saved[key] !== undefined){
					//console.log('key not null');
					//console.log(saved[key]);
					return saved[key];
				}else{
					//console.log('null');
					//console.log(saved[key]);
					return defaultValue;
				}
			}else{
				return defaultValue;
			}
			
		}else{
			//console.log('key null');
			return defaultValue;
		}
		
	},
	
	removeItem: function(key, menu){
		bngApi.engineLua('gdcallback.removeItem(' + bngApi.serializeToLua(key) + ',' + bngApi.serializeToLua(menu) + ')');
	},
	clearAll: function(){
		bngApi.engineLua('gdcallback.clearAll()');
	},
    set: function(key, value) {
      $window.localStorage[key] = value;
    },
    get: function(key, defaultValue) {
      return $window.localStorage[key] || defaultValue || false;
    },
    setObject: function(key, value) {
      $window.localStorage[key] = JSON.stringify(value);
    },
    getObject: function(key, defaultValue) {
      if($window.localStorage[key] != undefined){
          return JSON.parse($window.localStorage[key]);
      }else{
        return defaultValue;
      }
    },
    remove: function(key){
      $window.localStorage.removeItem(key);
    },
    clear: function(){
      $window.localStorage.clear();
    }
  }
}]);

app.factory('httpFactory', [
  '$http',
  '$q',
  function httpFactory($http, $q) {
    console.log('httpService Called');

    // interface
    var derbyService = {
      settings: {},
      getSettings: getSettings,
    };
    return derbyService;

    // implementation
    function getSettings(urls) {
      var def = $q.defer();
      var urlCalls = {};
      var key = 0;
      angular.forEach(urls, function(url) {
        urlCalls[key] = $http.get('gdcommon/' + url + '.json');
        key++;
      });
      $q.all(urlCalls).then(function successCallback(settings) {
        derbyService.settings = settings;
        def.resolve(settings);
        //console.log('derbyService Worked', settings);
      }),
        function errorCallback() {
          def.reject('Failed to get settings');
          //console.log('derbyService notworking');
        };
      return def.promise;
    }
  },
]);
app.controller('derbyController', [
  '$q',
  '$http',
  'httpFactory',
  '$log',
  '$scope',
  '$localstorage',
  'gamepadNav',
  function derbyController($q, $http, httpFactory, $log, $scope,$localstorage,gamepadNav) {
    console.log('derbyController Called');
	
	var crossfireEnabled = gamepadNav.crossfireEnabled();
	var gamepadNavEnabled = gamepadNav.gamepadNavEnabled();
	var spatialNavEnabled = gamepadNav.spatialNavEnabled();
	gamepadNav.enableCrossfire(false);
	gamepadNav.enableGamepadNav(false);
	gamepadNav.enableSpatialNav(true);
	
    var vm = this;
	vm.urls = ['menus','tracks'];
    vm.menus = {};
    vm.tracks = {};
	var track = {};
    vm.selected = {};
	vm.heading = {};
	vm.group = {};
	vm.levels = {};
	var	saved = {};
	
	vm.groupList = function(list) {
		var groupArray = [];
		angular.forEach(list, function (item, idx) {
			if (groupArray.indexOf(item.type) == -1) {
				//console.log(item.type)
				groupArray.push(item.type);
			}
		});
		//console.log(groupArray)
		vm.group = groupArray.sort();
	};
	var loadLogo = function () {
			var filler = document.getElementsByClassName('filler')[0].style
		//filler.backgroundImage = "linear-gradient(to bottom, red 0%,blue 100%),url('modules/gdcommon/background.jpg')";
		filler.backgroundImage = "linear-gradient(to bottom, red 0%,blue 100%),url('modules/gdcommon/gdcommonbg.jpg')";
		filler.backgroundSize = "100%";
		filler.backgroundRepeat = "no-repeat";
		filler.backgroundPosition = "center";
		filler.backgroundBlendMode = "soft-light";
	};
	var getDesc = function () {
		bngApi.engineLua('gdcallback.getDesc()',function (desc) {
			$scope.$evalAsync(function () {
				vm.heading.name = "";
				vm.heading.desc = "";
				if (desc) {
					console.log(desc.name);
					if (desc.found === 2) {
						loadLogo();
					};
					vm.heading = desc;
				}else{
					vm.heading.name = "Demolition Derby - The Fair Grounds v10.5.8";
					vm.heading.desc = "Race'em or Wreck'em";
				}
			});
		});
	};
	var loadSaved = function () {
		bngApi.engineLua('gdcallback.loadObject()',function (save) {
			$scope.$evalAsync(function () {
				saved = save
			});
		});
	};

   
	vm.update = function (refresh) {
		if (refresh === 1){
			var parts = $localstorage.loadObject("parts","",saved)
			if (parts) {
				vm.selected.parts = parts
			}
		};
		
		if (refresh === 1){
			
			$localstorage.saveTable(vm.selected)//We need to update the classes lua file before the getData call. 
			bngApi.engineLua('gdcallback.getData("' + vm.selected.map + '")',function (response) {
				$scope.$evalAsync(function () {	
					var config2 = response.config2;
					var model = response.model;
					var map = vm.selected.map;
					var lastmap = $localstorage.loadObject('map',map,saved);
					var flevels = response.flevels[map];
					vm.selected.fmap = flevels;
					if (response.tracks !== undefined) {
						vm.tracks = response.tracks.arenas[flevels];
						var defaultArena = $localstorage.loadObject('arena',response.tracks.arenas[flevels].main.arena.default,saved);	
						if (lastmap !== map) {			
							defaultArena =  response.tracks.arenas[flevels].main.arena.default;
						};							
						vm.selected.arena = defaultArena;
					}else{
						if (track.arenas[flevels]) {
							vm.tracks = track.arenas[flevels];
							var defaultArena = $localstorage.loadObject('arena',track.arenas[flevels].main.arena.default,saved);
							if (lastmap !== map) {
								defaultArena = track.arenas[flevels].main.arena.default;
							};
							//console.log(defaultArena);
							vm.selected.arena = defaultArena;
						}else{
							vm.tracks = track.arenas['common'];
							//console.log('common');
							vm.selected.arena = track.arenas['common'].main.arena.default;
						};	
					};
					vm.menus.main.config2.listob = config2[0].listob;
					vm.selected.config2 = config2[0].default;
					vm.menus.main.vehicles.listg = model;
					vm.selected.vehicles = vm.menus.main.vehicles.default;
				});
			});
		};
		$localstorage.saveTable(vm.selected)
    };
	 var getSettings = function() {
      httpFactory.getSettings(vm.urls).then(
        function(settings) {
			//$localstorage.clear();
			vm.menus = settings[0].data;
			track = settings[1].data;
			vm.tracks = track.arenas['common'];
			bngApi.engineLua('gdcallback.getData()',function (mapData) {
				$scope.$evalAsync(function () {
					var map = mapData.map;  //currentmap starting with gdcommon  is the current map that is loaded
					var levels = mapData.levels; //List of Maps or Levels the entire list of maps in the game
					//vm.selected.fmap = mapData.flevels[levels];  //current level in original format with special chars
					vm.menus.main.map.listob = levels;
					if (map !== 'GDCOMMON') {
						vm.menus.main.map.default = map
						//if (mapData.tracks !== undefined) {
						//var customTracks = mapData.tracks;
							//if (customTracks) {
								//track.main.arena.listg.concat(customTracks.main.arena.listg);
							//	vm.tracks = customTracks;
							//};
						//};
					};
		
					//var model = $localstorage.loadObject('model', model,saved);
					//$localstorage.removeItem('vehicles');
					//set the selection from saved selections or load the default selection
					vm.urls.forEach(function (menu) {
						angular.forEach(vm[menu].main,function(value, key) {
							if (value.name === 'Arena') {
								vm.selected['laps'] = $localstorage.loadObject('laps',value.laps.default,saved);
							};
							vm.selected[key] = $localstorage.loadObject(key,value.default,saved);
						});
					});
					
					angular.forEach(vm.menus.settings,function(value, key) {
						var settings = $localstorage.loadObject(key,value.default,saved);
						switch (settings) {
							case 'false':
								vm.selected[key] = false;
								break;
							case 'true':
								vm.selected[key] = true;
								break;
							default:
								vm.selected[key] = settings;
						}
					});
					
					//console.log('selected', vm.selected);
					vm.update(1);
				});	
			});	
        },
      );
    };	
	getDesc();
	loadSaved();
	getSettings();
	
	vm.clear = function(menu){
		$localstorage.removeItem(menu,'parts')
		vm.selected.parts[menu] = {}
	};
   vm.section = function(index) {
      var acc = document.getElementsByClassName('accordion')
      acc[index].classList.toggle("active");
      acc[index].nextElementSibling.classList.toggle("show");
    };
    vm.tab = 1;
    vm.setTab = function(newTab) {
      vm.tab = newTab;
	  if (vm.tab === 3) {
		  bngApi.engineLua('gdcallback.getParts()',function (response) {
			$scope.$evalAsync(function () {
				var errors = response.parts[0].msg;	
				vm.menus.parts = {};				
			if (errors) {
				
				vm.menus.parts.partsGroup = errors;
			}else{
				var partsGroup = response.parts[0].partsGroup;
				var partsMap = response.parts[0].partsMap;
				var slotType = response.parts[0].slotType;
				vm.menus.parts.partsGroup = partsGroup;
				vm.menus.parts.partsMap = partsMap;
				vm.menus.parts.slotType = slotType;
			};
			});
		});
	  }
	  
    };
    vm.isSet = function(tabNum) {
      return vm.tab === tabNum;
    };
	$scope.$on('$destroy', function() {
    bngApi.engineLua('WinInput.setForwardRawEvents(false);');
    bngApi.engineLua('WinInput.setForwardFilteredEvents(false);');
    gamepadNav.enableCrossfire(crossfireEnabled); // use old value from before opening the menu here
    gamepadNav.enableGamepadNav(gamepadNavEnabled); // use old value here
    gamepadNav.enableSpatialNav(spatialNavEnabled); // use old value here
    $scope.$emit('ShowApps', true);
  });
  },
  
]);
app.directive('derbyMainTab', function() {
  return {
    restrict: 'E',
    replace: true,
    templateUrl: 'gdcommon/main.html',
  };
});
app.directive('derbySettingsTab', function() {
  return {
    restrict: 'E',
    replace: true,
    templateUrl: 'gdcommon/settings.html',
    scope: {
      settings: '=',
    },
  };
});
app.directive('derbyPartsTab', function() {
  return {
    restrict: 'E',
    replace: true,
    templateUrl: 'gdcommon/parts.html',
  };
});
app.directive('derbyOptionsGroup', function() {
  return {
    restrict: 'E',
    replace: true,
    template:
      `<md-select class="dropdown" id="menu.name" ng-model="c.selected[key]" ng-model-options="{trackBy: '$value.name'}"
	  aria-label="menu.name" ng-change="c.update(menu.refresh)">
	  <md-optgroup label="{{groupName}}" ng-repeat="groupName in c.group">
	  <md-option ng-value="{{list}}"  ng-repeat="list in menu.listg | filter: {type:groupName} | orderBy:'+name' ">{{list.name}}
	  </md-option>
	  </md-optgroup> 
	  </md-select>`
  };
});
app.directive('derbyOptionsOrderBy', function() {
  return {
    restrict: 'E',
    replace: true,
    template:
      `<md-select class="dropdown" id="menu.name" ng-model="c.selected[key]" aria-label="menu.name" ng-change="c.update(menu.refresh)"> 
	  <md-option ng-repeat="list in menu.listob| orderBy: list">{{list}}</md-option> </md-select>`,
  };
});
app.directive('derbyOptions', function() {
  return {
    restrict: 'E',
    replace: true,
    template:
      `<md-select class="dropdown" id="menu.name" ng-model="c.selected[key]" aria-label="menu.name" ng-change="c.update(menu.refresh)"> 
	  <md-option ng-repeat="list in menu.listng">{{list}}</md-option></md-select>`,
  };
});
app.directive('derbyTooltip', function() {
  return {
    restrict: 'E',
    replace: true,
    template:
      `<span class="tooltip"> <img src="help.png"> <div class="tooltiptext"> <div ng-repeat="tips in menu.tooltip">{{tips}} </div> </div> </span>`,
  };
});
app.config([
    '$stateProvider',
    '$urlRouterProvider',
    function ($stateProvider, $urlRouterProvider) {
      $stateProvider
      .state('gdEndScreen', {
        url: '/gdEndScreen/end',
		 params: {
			endData: {},
			drivers: {},
			headers: {},
		},	
		templateUrl: 'modules/gdcommon/gdEndScreen/end.html',
        controller: 'gdEndController',
		 
      })
	  $urlRouterProvider.otherwise('/gdEndScreen/end');
    },
]);
app.controller('gdEndController', ['logger', '$scope', '$sce', 'Utils', '$state', '$stateParams', '$timeout', '$filter', 'gamepadNav',
function (logger, $scope, $sce, Utils, $state, $stateParams, $timeout, $filter, gamepadNav) {
  var crossfireEnabled = gamepadNav.crossfireEnabled();
  var gamepadNavEnabled = gamepadNav.gamepadNavEnabled();
  var spatialNavEnabled = gamepadNav.spatialNavEnabled();
  gamepadNav.enableCrossfire(false);
  gamepadNav.enableGamepadNav(false);
  gamepadNav.enableSpatialNav(true);
  Utils.waitForCefAndAngular(() => {
    bngApi.engineLua("extensions.core_input_bindings.menuActive(true);");
  });

  var data = $stateParams.endData;
  
  // debugger;
	if (data.options === undefined){
		 bngApi.engineLua("gdscenarios.endScreen();");
		 return
	};
 
	$scope.data = data; 
	var title = data.options.map + ' - ' + data.options.arena.name;
	
	//remove _ in titles
	$scope.eventTitle = title.replace(/_/g, ' ');
	
	$scope.drivers = $stateParams.drivers;
	$scope.headers = $stateParams.headers;
	
	$scope.filter = function(value, filter) {
   if(filter)
    {
      return $filter(filter)(value);
    }{
      
      return value
      }
  };
	// column to sort
	$scope.column = 'place';
	 
	 // sort ordering (Ascending or Descending). Set true for desending
	$scope.reverse = false; 
	
	// called on header click
	$scope.sortColumn = function(col){
	$scope.column = col;
	if($scope.reverse){
		$scope.reverse = false;
		$scope.reverseclass = 'arrow-up';
	}else{
		$scope.reverse = true;
		$scope.reverseclass = 'arrow-down';
		}
	};
	
	// remove and change class
	$scope.sortClass = function(col){
	if($scope.column == col ){
		if($scope.reverse){
			return 'arrow-down'; 
		}else{
			return 'arrow-up';
		}
	}else{
		return '';
	}
	}; 
 
  //$scope.BBToHtml = function (code) {
  //  console.log(code);
   // return $sce.trustAsHtml(Utils.parseBBCode(code));
  //};
	$scope.$emit('ShowApps', false);  // hide apps when time trials is open
  $scope.activeButton = $scope.data.buttons.length - 1;

  for (var i = $scope.data.buttons.length - 1; i >= 0; i--) {
    if ($scope.data.buttons[i].active) {
      $scope.activeButton = i;
    }
  }

  var cmds = {
    openMenu: function () {
      $scope.$emit('MenuToggle', true);
	   var luaCmd = 'scenario_scenarios.changeState("post")';
        bngApi.engineLua(luaCmd);
    },
    openScenarios:function(){
      $scope.$emit('MenuToggle', true);  
      setTimeout(function() {$state.go('menu.scenarios');});
    },
  };

  function execHelper (cmd) {
    var luaCmd = cmds[cmd] === undefined;
    return {
      func: luaCmd ? () => bngApi.engineLua(cmd) : cmds[cmd],
      exits: luaCmd
    };
  }

  $scope.executeCmd = function (cmdStr, showLoadingScreen) {
    console.log(cmdStr)
    var cmd = execHelper(cmdStr);

    if (cmd.exits) {
      $scope.$emit('StickyState', null);
    }
    cmd.func();
  };

  $scope.getBarSound = function (i) {
    if (i === 0) {
      return {animationstart: 'event:>UI>Scenario End Counting'};
    }
    return {};
  };


 

  $scope.$on('$destroy', () => {
    // CampaignResults.stats = $scope.data;
    $scope.$emit('ShowApps', true);
    gamepadNav.enableCrossfire(crossfireEnabled); // use old value from before opening the menu here
    gamepadNav.enableGamepadNav(gamepadNavEnabled); // use old value here
    gamepadNav.enableSpatialNav(spatialNavEnabled); // use old value here
    bngApi.engineLua("extensions.core_input_bindings.menuActive(false);");
  });

  setTimeout(function() {
    $scope.$emit('StickyState', $state.$current.name);
    $scope.$emit('ShowApps', false);
  });

  //$scope.toggleDetail = function () {
   // $scope.$evalAsync(() => {
    //   $scope.detailed = !$scope.detailed;
    //});
  //};
}]);